"""Library for interfacing with an IsatData Pro modem for satellite IoT."""

# Workaround for legacy async client
from idpmodem.asyncio import atcommand_async
from idpmodem.asyncio.atcommand_async import IdpModemAsyncioClient

__all__ = ['atcommand_async', 'IdpModemAsyncioClient']



# -*- coding: utf-8 -*-
"""IsatData Pro modem constants.

This module provides mapping of constants used within an IDP modem.

"""

from enum import Enum, IntEnum, IntFlag


class MessagePriority(IntEnum):
    NONE = 0
    HIGH = 1
    MEDH = 2
    MEDL = 3
    LOW = 4


class DataFormat(IntEnum):
    TEXT = 1
    HEX = 2
    BASE64 = 3


CONTROL_STATES = {
    0: 'Stopped',
    1: 'Waiting for GNSS fix',
    2: 'Starting search',
    3: 'Beam search',
    4: 'Beam found',
    5: 'Beam acquired',
    6: 'Beam switch in progress',
    7: 'Registration in progress',
    8: 'Receive only',
    9: 'Downloading Bulletin Board',
    10: 'Active',
    11: 'Blocked',
    12: 'Confirm previously registered beam',
    13: 'Confirm requested beam',
    14: 'Connect to confirmed beam'
}


class SatlliteControlState(IntEnum):
    STOPPED = 0
    GNSS_WAIT = 1
    SEARCH_START = 2
    BEAM_SEARCH = 3
    BEAM_FOUND = 4
    BEAM_ACQUIRED = 5
    BEAM_SWITCH = 6
    REGISTERING = 7
    RECEIVE_ONLY = 8
    BB_DOWNLOAD = 9
    ACTIVE = 10
    BLOCKED = 11
    CONFIRM_PREVIOUS_BEAM = 12
    CONFIRM_REQUESTED_BEAM = 13
    CONNECT_CONFIRMED_BEAM = 14


class BeamSearchState(IntEnum):
    IDLE = 0
    SEARCH_ANY_TRAFFIC = 1
    SEARCH_LAST_TRAFFIC = 2
    RESERVED = 3
    SEARCH_NEW_TRAFFIC = 4
    SEARCH_BULLETIN_BOARD = 5
    DELAY_TRAFFIC_SEARCH = 6


class MessageState(IntEnum):
    UNAVAILABLE = 0
    RX_PENDING = 1
    RX_COMPLETE = 2
    RX_RETRIEVED = 3
    TX_READY = 4
    TX_SENDING = 5
    TX_COMPLETE = 6
    TX_FAILED = 7
    TX_CANCELLED = 8


AT_ERROR_CODES = {
    0: 'OK',
    4: 'ERROR',
    100: 'ERR_INVALID_CRC_SEQUENCE',
    101: 'ERR_UNKNOWN_COMMAND',
    102: 'ERR_INVALID_COMMAND_PARAMETERS',
    103: 'ERR_MESSAGE_LENGTH_EXCEEDS_FORMAT_SIZE',
    104: 'ERR_RESERVED_104',
    105: 'ERR_SYSTEM_ERROR',
    106: 'ERR_QUEUE_INSUFFICIENT_RESOURCES',
    107: 'ERR_MESSAGE_NAME_ALREADY_IN_USE',
    108: 'ERR_TIMEOUT_OCCURRED',
    109: 'ERR_UNAVAILABLE',
    110: 'ERR_RESERVED_110',
    111: 'ERR_RESERVED_111',
    112: 'ERR_ATTEMPT_TO_WRITE_READ_ONLY_PARAMETER'
}


WAKEUP_PERIODS = {
    0: 'SECONDS_5',
    1: 'SECONDS_30',
    2: 'MINUTES_1',
    3: 'MINUTES_3',
    4: 'MINUTES_10',
    5: 'MINUTES_30',
    6: 'MINUTES_60',
    7: 'MINUTES_2',
    8: 'MINUTES_5',
    9: 'MINUTES_15',
    10: 'MINUTES_20'
}

POWER_MODES = {
    0: 'MOBILE_POWERED',
    1: 'FIXED_POWERED',
    2: 'MOBILE_BATTERY',
    3: 'FIXED_BATTERY',
    4: 'MOBILE_MINIMAL',
    5: 'MOBILE_PARKED'
}


class IdpEnum(IntEnum):
    @classmethod
    def is_valid(cls, value: int) -> bool:
        """True if the value exists in the enumeration."""
        return value in [v.value for v in cls.__members__.values()]


class AtErrorCode(IdpEnum):
    OK = 0
    ERROR = 4
    INVALID_CRC = 100
    UNKNOWN_COMMAND = 101
    INVALID_COMMAND_PARAMETERS = 102
    MESSAGE_LENGTH_EXCEEDS_FORMAT_SIZE = 103
    RESERVED_104 = 104
    SYSTEM_ERROR = 105
    QUEUE_INSUFFICIENT_RESOURCES = 106
    DUPLICATE_MESSAGE_NAME = 107
    GNSS_TIMEOUT = 108
    MESSAGE_UNAVAILABLE = 109
    RESERVED_110 = 110
    RESERVED_111 = 111
    READ_ONLY_PARAMETER = 112


class PowerMode(IdpEnum):
    MOBILE_POWERED = 0
    FIXED_POWERED = 1
    MOBILE_BATTERY = 2
    FIXED_BATTERY = 3
    MOBILE_MINIMAL = 4
    MOBILE_PARKED = 5

    def gnss_refresh_hours(self):
        """The minimum GNSS refresh interval [hours]."""
        if self.value == 0:
            return 3
        if self.value == 1:
            return 24
        if self.value == 2:
            return 6
        if self.value == 3:
            return 14 * 24
        if self.value == 4:
            return 12
        if self.value == 5:
            return 24
    
    def transmit_lifetime_seconds(self):
        """The maximum duration of a message in the transmit queue [seconds]."""
        if self.value in [0, 1]:
            return 3 * 3600
        return  3 * 60
    
    def beam_search_interval_seconds(self):
        """The minimum time between background beam searches [seconds]."""
        if self.value in [0, 2, 4]:
            return 20 * 60
        return 60 * 60

    def short_term_blockage_seconds(self):
        """The time in blockage before initiating a beam search [seconds]."""
        if self.value in [0, 1]:
            return 5 * 60
        if self.value in [2, 3, 4]:
            return 20 * 60
        return 15 * 60
    
    def beam_search_maximum_seconds(self):
        """The maximum backoff interval between searches when blocked [seconds].
        """
        if self.value in [0, 1]:
            return 0
        return 1600 * 60


class WakeupPeriod(IdpEnum):
    SECONDS_5 = 0
    SECONDS_30 = 1
    MINUTES_1 = 2
    MINUTES_3 = 3
    MINUTES_10 = 4
    MINUTES_30 = 5
    MINUTES_60 = 6
    MINUTES_2 = 7
    MINUTES_5 = 8
    MINUTES_15 = 9
    MINUTES_20 = 10

    def seconds(self):
        value = int(self.name.split('_'))[1]
        if self.name.startswith('MINUTES'):
            return value * 60
        return value


class GnssMode(IdpEnum):
    GPS = 0
    GLONASS = 1
    BEIDOU = 2
    GALILEO = 3
    GPS_GLONASS = 10
    GPS_BEIDOU = 11
    GLONASS_BEIDOU = 12
    GPS_GALILEO = 13
    GLONASS_GALILEO = 14
    BEIDOU_GALILEO = 15


# Dynamic Platform Model - use with caution
GNSS_DPM_MODES = {
    0: 'PORTABLE',
    2: 'STATIONARY',
    3: 'PEDESTRIAN',
    4: 'AUTOMOTIVE',
    5: 'SEA',
    6: 'AIR_1G',
    7: 'AIR_2G',
    8: 'AIR_4G'
}


class SignalLevelRegional(Enum):
    """Qualitative descriptors for SNR/CN0 values for a IDP Regional Beam.
    
    BARS_n: *n* is a scale from 0..5 to be used as greaterThan threshold
    NONE, MARGINAL, GOOD: a scale to be used as greaterOrEqual threshold

    """
    BARS_0 = 0
    BARS_1 = 37.0
    BARS_2 = 39.0
    BARS_3 = 41.0
    BARS_4 = 43.0
    BARS_5 = 45.5
    INVALID = 55.0


class SignalQuality(IntEnum):
    NONE = 0
    WEAK = 1
    LOW = 2
    MID = 3
    GOOD = 4
    STRONG = 5
    WARNING = 6


class EventNotification(IntFlag):
    GNSS_FIX_NEW = 0b000000000001
    MESSAGE_MT_RECEIVED = 0b000000000010
    MESSAGE_MO_COMPLETE = 0b000000000100
    NETWORK_REGISTERED = 0b000000001000
    MODEM_RESET_COMPLETE = 0b000000010000
    JAMMING_ANTENNA_CHANGE = 0b000000100000
    MODEM_RESET_PENDING = 0b000001000000
    WAKEUP_PERIOD_CHANGE = 0b000010000000
    UTC_TIME_SYNC = 0b000100000000
    GNSS_FIX_TIMEOUT = 0b001000000000
    EVENT_TRACE_CACHED = 0b010000000000
    NETWORK_PING_ACKNOWLEDGED = 0b100000000000


class TransmitterStatus(IdpEnum):
    RX_ONLY_NOT_REGISTERED = 4
    OK = 5
    SUSPENDED = 6
    MUTED = 7
    BLOCKED = 8


class EventTraceClass(IdpEnum):
    HARDWARE_FAULT = 1
    SYSTEM = 2
    SATELLITE = 3
    GNSS = 4
    MESSAGE = 5


class EventTraceSubclass(IdpEnum):
    """Base class for trace subclasses"""


class EventTraceSystem(EventTraceSubclass):
    RESET = 1
    LOW_POWER = 2
    SYSTEM_STATS = 3
    SATCOM_STATS = 4


class EventTraceSatellite(EventTraceSubclass):
    RX_STATUS = 1
    TX_STATUS = 2
    BEAM_CONNECT = 3
    BEAM_SEARCH_RESULT = 4
    GEO_ADJUST = 5
    EVENT_LOG = 6
    RX_METRICS_SUBFRAME = 16
    RX_METRICS_MINUTE = 17
    RX_METRICS_HOUR = 18
    RX_METRICS_DAY = 19
    TX_METRICS_SUBFRAME = 20
    TX_METRICS_MINUTE = 21
    TX_METRICS_HOUR = 22
    TX_METRICS_DAY = 23


class EventTraceGnss(EventTraceSubclass):
    FIX_STATS = 1
    DOPPLER = 4


class EventTraceMessage(EventTraceSubclass):
    RECEIVE_STATS = 1
    TRANSMIT_STATS = 2
    TRANSMIT_UTILITY = 3


class EventTrace:
    def __init__(self,
                 trace_class: EventTraceClass,
                 trace_subclass: EventTraceSubclass,
                 data: 'tuple[str, str|dict|IntEnum]') -> None:
        """Defines an event trace.
        
        Args:
            trace_class: The enumerated class
            trace_subclass: The enumerated subclass dependent on the class
            data: The set of (name, meta) where meta defines either a
                data type (e.g. `uint`) or a mapping (`dict` or `IntEnum`)
        """
        self.trace_class: EventTraceClass = trace_class
        self.trace_subclass: EventTraceSubclass = trace_subclass
        self.data: 'tuple[str, str|dict|IntEnum]' = data


EVENT_TRACE_SATELLITE_GENERAL = EventTrace(
    trace_class=EventTraceClass.SATELLITE,
    trace_subclass=EventTraceSatellite.RX_STATUS,
    data=(
        ('subframe_number', 'uint'),
        ('traffic_vcid', 'uint'),
        ('configuration_id', 'uint'),
        ('beam_number', 'uint'),
        ('reserved04', 'uint'),
        ('tx_access_sip', 'uint'),
        ('tx_access_operator', 'uint'),
        ('tx_access_user', 'uint'),
        ('tx_suspend_flags', {
            0x1: 'BEAM_REGISTRATION',
            0x2: 'BEAM_SWITCH',
            0x4: 'RESERVED',
            0x8: 'BLOCKED',
        }),
        ('tx_messages_active', 'uint'),
        ('tx_messages_total', 'uint'),
        ('tx_packets_active', 'uint'),
        ('tx_state', {
            0: 'ACTIVE',
            1: 'SUSPENDING',
            2: 'SUSPENDED_PENDING_GNSS'}),
        ('active_rx_messages', 'uint'),
        ('beamswitch_averaging_window', 'uint'),
        ('beamswitch_averaging_count', 'uint'),
        ('c_n_x100', 'uint'),
        ('beamsample_threshold', 'uint'),
        ('beamsample_timer', 'uint'),
        ('flags', {
            0x1: 'REGISTERED',
            0x2: 'SENDING_BEAM_REGISTRATION',
            0x10: 'BEAM_SEARCH',
            0x20: 'BEAM_SAMPLE_REQUIRED',
            0x40: 'BEAM_SWITCH_PENDING',
            0x100: 'GNSS_VALID',
            0x200: 'GNSS_REQUIRED',
            0x400: 'GNSS_PENDING',
        }),
        ('gnss_state_timer', 'uint'),
        ('reserved21', 'uint'),
        ('satellite_control_state', SatlliteControlState),
        ('beam_search_state', BeamSearchState),
    )
)


EVENT_TRACES = (
    EVENT_TRACE_SATELLITE_GENERAL,
)


GEOBEAMS = {
    1: 'AMER RB1',
    2: 'AMER RB2',
    3: 'AMER RB3',
    4: 'AMER RB4',
    5: 'AMER RB5',
    6: 'AMER RB6',
    7: 'AMER RB7',
    8: 'AMER RB8',
    9: 'AMER RB9',
    10: 'AMER RB10',
    11: 'AMER RB11',
    12: 'AMER RB12',
    13: 'AMER RB13',
    14: 'AMER RB14',
    15: 'AMER RB15',
    16: 'AMER RB16',
    17: 'AMER RB17',
    18: 'AMER RB18',
    19: 'AMER RB19',
    61: 'AORW SC',
    21: 'EMEA RB1',
    22: 'EMEA RB2',
    23: 'EMEA RB3',
    24: 'EMEA RB4',
    25: 'EMEA RB5',
    26: 'EMEA RB6',
    27: 'EMEA RB7',
    28: 'EMEA RB8',
    29: 'EMEA RB9',
    30: 'EMEA RB10',
    31: 'EMEA RB11',
    32: 'EMEA RB12',
    33: 'EMEA RB13',
    34: 'EMEA RB14',
    35: 'EMEA RB15',
    36: 'EMEA RB16',
    37: 'EMEA RB17',
    38: 'EMEA RB18',
    39: 'EMEA RB19',
    41: 'APAC RB1',
    42: 'APAC RB2',
    43: 'APAC RB3',
    44: 'APAC RB4',
    45: 'APAC RB5',
    46: 'APAC RB6',
    47: 'APAC RB7',
    48: 'APAC RB8',
    49: 'APAC RB9',
    50: 'APAC RB10',
    51: 'APAC RB11',
    52: 'APAC RB12',
    53: 'APAC RB13',
    54: 'APAC RB14',
    55: 'APAC RB15',
    56: 'APAC RB16',
    57: 'APAC RB17',
    58: 'APAC RB18',
    59: 'APAC RB19',
    90: 'MEAS RB10',
    91: 'MEAS RB11',
    92: 'MEAS RB12',
    93: 'MEAS RB15',
}


class GeoBeam(IdpEnum):
    GLOBAL_BB = 0
    AMER_RB1 = 1
    AMER_RB2 = 2
    AMER_RB3 = 3
    AMER_RB4 = 4
    AMER_RB5 = 5
    AMER_RB6 = 6
    AMER_RB7 = 7
    AMER_RB8 = 8
    AMER_RB9 = 9
    AMER_RB10 = 10
    AMER_RB11 = 11
    AMER_RB12 = 12
    AMER_RB13 = 13
    AMER_RB14 = 14
    AMER_RB15 = 15
    AMER_RB16 = 16
    AMER_RB17 = 17
    AMER_RB18 = 18
    AMER_RB19 = 19
    AORW_SC = 61
    EMEA_RB1 = 21
    EMEA_RB2 = 22
    EMEA_RB3 = 23
    EMEA_RB4 = 24
    EMEA_RB5 = 25
    EMEA_RB6 = 26
    EMEA_RB7 = 27
    EMEA_RB8 = 28
    EMEA_RB9 = 29
    EMEA_RB10 = 30
    EMEA_RB11 = 31
    EMEA_RB12 = 32
    EMEA_RB13 = 33
    EMEA_RB14 = 34
    EMEA_RB15 = 35
    EMEA_RB16 = 36
    EMEA_RB17 = 37
    EMEA_RB18 = 38
    EMEA_RB19 = 39
    APAC_RB1 = 41
    APAC_RB2 = 42
    APAC_RB3 = 43
    APAC_RB4 = 44
    APAC_RB5 = 45
    APAC_RB6 = 46
    APAC_RB7 = 47
    APAC_RB8 = 48
    APAC_RB9 = 49
    APAC_RB10 = 50
    APAC_RB11 = 51
    APAC_RB12 = 52
    APAC_RB13 = 53
    APAC_RB14 = 54
    APAC_RB15 = 55
    APAC_RB16 = 56
    APAC_RB17 = 57
    APAC_RB18 = 58
    APAC_RB19 = 59
    MEAS_RB10 = 90
    MEAS_RB11 = 91
    MEAS_RB12 = 92
    MEAS_RB15 = 93

    def satellite(self):
        return self.name.split('_')[0]

    def beam(self):
        return self.name.split('_')[1]
    
    def id(self):
        return self.value



# -*- coding: utf-8 -*-
"""AT command protocol (asyncio) for Inmarsat IDP satellite messaging modems.

This module provides an async serial interface that sends and receives 
AT commands, decoding/abstracting typically used operations.
Based on the AioSerial package.

**WARNING**: This module needs significant refactoring and may be unstable.

Example usage::

    import asyncio
    import idpmodem

    mymodem = idpmodem.IdpModemAsyncioClient()
    initialized = asyncio.run(mymodem.initialize(crc=True))

"""

import logging
from asyncio import (AbstractEventLoop, TimeoutError, gather, get_running_loop,
                     run_coroutine_threadsafe, set_event_loop, sleep, wait_for,
                     wrap_future)
from base64 import b64decode, b64encode
from collections import OrderedDict
from glob import glob
from threading import Event, current_thread
from time import time
from typing import Tuple, Union

from aioserial import AioSerial
from idpmodem.aterror import (AtCrcConfigError, AtCrcError, AtException,
                              AtGnssTimeout, AtTimeout)
from idpmodem.constants import (AT_ERROR_CODES, CONTROL_STATES, POWER_MODES,
                                WAKEUP_PERIODS, BeamSearchState, DataFormat,
                                MessageState, TransmitterStatus)
from idpmodem.crcxmodem import get_crc, validate_crc
from idpmodem.location import Location, location_from_nmea

_log = logging.getLogger(__name__)

BAUDRATES = [2400, 4800, 9600, 19200, 38400, 57600, 115200]

NOTIFICATION_BITMASK = (
    'gnss_fix_new',
    'message_mt_received',
    'message_mo_complete',
    'network_registered',
    'modem_reset',
    'jamming_antenna_change',
    'modem_reset_pending',
    'wakeup_period_changed',
    'utc_time_set',
    'gnss_fix_timeout',
    'event_cached',
    'network_ping_acknowledged'
)


def _printable(string: str) -> str:
    """Used for visualizing non-printable AT characters."""
    return string.replace('\r', '<cr>').replace('\n', '<lf>')


def _serial_asyncio_lost_bytes(response: str) -> bool:
    """Used to capture errors in serial_asyncio.
    
    Deprecated.
    """
    if ('AT' in response or '\r\r' in response):
        return True
    return False


def _to_signed32(n):
    """Converts an integer to signed 32-bit format."""
    n = n & 0xffffffff
    return (n ^ 0x80000000) - 0x80000000


def _notifications_dict(sreg_value: int = None) -> OrderedDict:
    """Returns an OrderedDictionary as an abstracted bitmask of notifications.
    
    Args:
        sreg_value: (optional) the integer value stored in S88 or S89
    
    Returns:
        ordered dictionary corresponding to bitmask
    """
    template = OrderedDict([
        (bit, False) for bit in NOTIFICATION_BITMASK])
    if sreg_value is not None:
        bitmask = bin(int(sreg_value))[2:]
        if len(bitmask) > len(template):
            bitmask = bitmask[:len(template) - 1]
        while len(bitmask) < len(template):
            bitmask = '0' + bitmask
        i = 0
        for key in reversed(template):
            template[key] = True if bitmask[i] == '1' else False
            i += 1
    return template


class IdpModemAsyncioClient:
    """A satellite IoT messaging modem on Inmarsat's IsatData Pro service.

    **WARNING**: Deprecated

    Attributes:
        port: The serial port name e.g. `/dev/ttyUSB0`
        baudrate: The baudrate of the serial port e.g. `9600`
        crc: A boolean used if CRC-16 is enabled for long serial cables
        loop: The asyncio event loop (uses default if not provided)

    """

    def __init__(self,
                 port: str = '/dev/ttyUSB0',
                 baudrate: int = 9600,
                 loop: AbstractEventLoop = None,
                 log_verbose: bool = False,
                 ):
        """Initializes the class.
        
        Args:
            port: The serial port name e.g. `/dev/ttyUSB0`
            baudrate: The serial port baudrate
            crc: enables CRC-16 for long serial cables
            loop: (optional) external asyncio event loop to use
            logger: (optional) external logger to use
            log_level: Level for the logger to record

        """
        self._verbose = log_verbose
        self.port = port
        self.baudrate = baudrate
        self.crc = None
        self.loop = loop
        self._thread = current_thread()
        self._event = Event()
        self._serial = None
        self._pending_command = None
        self._pending_command_time = None
        self._retry_count = 0
        self._serial_async_error_count = 0

    @property
    def port(self):
        return self._port
    
    @port.setter
    def port(self, value):
        valid = len(glob(value)) == 1
        if not valid:
            err_msg = 'Serial port {} not found'.format(value)
            _log.error(err_msg)
            raise ValueError(err_msg)
        self._port = value

    @property
    def baudrate(self):
        return self._baudrate
    
    @baudrate.setter
    def baudrate(self, value):
        if value not in BAUDRATES:
            raise ValueError('Unsupported baudrate {}'.format(value))
        self._baudrate = value

    def _handle_at_error(self,
                         at_command: str,
                         err_code: Union[str, int],
                         return_value: any = None) -> any:
        """Manages log and/or raising errors.
        
        Args:
            at_command: The command that experienced an error
            err_code: The error code received
            return_value: The value to return after logging
        
        Raises:
            Re-raises the exceptions

        """
        error_str = AT_ERROR_CODES[int(err_code)]
        _log.error("{} Exception: {}".format(at_command, error_str))
        if return_value is None:
            raise AtException(error_str)
        return return_value

    async def _send(self, data: str) -> str:
        """Coroutine encodes and sends an AT command.
        
        Args:
            writer: A serial_asyncio writer
            data: An AT command string
        
        Returns:
            A string with the original data.
        """
        if self.crc:
            data = get_crc(data)
        self._pending_command = data
        to_send = self._pending_command + '\r'
        if self._verbose:
            _log.debug('Sending {}'.format(_printable(to_send)))
        self._pending_command_time = time()
        await self._serial.write_async(to_send.encode())
        return data

    async def _recv(self, timeout: int = 5) -> list:
        """Coroutine receives and decodes data from the serial port.

        Parsing stops when 'OK' or 'ERROR' is found.
        
        Args:
            reader: A serial_asyncio reader

        Returns:
            A list of response strings with empty lines removed.
        
        Raises:
            AtTimeout if the response timed out.

        """
        CRC_DELAY = 1   #: seconds after response body
        response = []
        verbose_response = ''
        msg = ''
        try:
            while True:
                chars = (await wait_for(
                    self._serial.read_until_async(b'\r\n'),
                    timeout=timeout)).decode()
                msg += chars
                verbose_response += chars
                if msg.endswith('\r\n'):
                    if self._verbose:
                        _log.debug('Processing {}'.format(_printable(msg)))
                    msg = msg.strip()
                    if msg != self._pending_command:
                        if msg != '':
                            # empty lines are not included in response list
                            # but are preserved in verbose_response for CRC
                            response.append(msg)
                    else:
                        # remove echo for possible CRC calculation
                        echo = self._pending_command + '\r'
                        if self._verbose:
                            _log.debug(f'Removing echo {_printable(echo)}')
                        verbose_response = verbose_response.replace(echo, '')
                    if msg in ['OK', 'ERROR']:
                        try:
                            response_crc = (await wait_for(
                                self._serial.read_until_async(b'\r\n'),
                                timeout=CRC_DELAY)).decode()
                            if response_crc:
                                response_crc = response_crc.strip()
                                if _serial_asyncio_lost_bytes(verbose_response):
                                    self._serial_async_error_count += 1
                                if not validate_crc(response=verbose_response,
                                                    candidate=response_crc):
                                    err_msg = '{} CRC error for {}'.format(
                                        response_crc,
                                        _printable(verbose_response))
                                    _log.error(err_msg)
                                    raise AtCrcError(err_msg)
                                elif self._verbose:
                                    _log.debug('CRC {} ok for {}'.format(
                                        response_crc,
                                        _printable(verbose_response)))
                                if not self.crc:
                                    # raise AtCrcConfigError('CRC found but unexpected') #: new <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                                    self.crc = True
                        except TimeoutError:
                            if self.crc:
                                raise AtCrcConfigError('CRC expected but not found')   #: new <<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<
                            self.crc = False
                        break
                    msg = ''
        except TimeoutError:
            timeout_time = time() - self._pending_command_time
            err = ('AT timeout {} after {} seconds ({}s after command)'.format(
                self._pending_command, timeout, timeout_time))
            raise AtTimeout(err)
        return response

    async def command(self,
                      at_command: str,
                      timeout: int = 5,
                      retries: int = 0) -> list:
        """Submits an AT command and returns the response asynchronously.
        
        Proxies a private function to allow for multi-threaded operation.

        Args:
            at_command: The AT command string
            timeout: The maximum time in seconds to await a response.
            retries: Optional number of additional attempts on failure.
        
        Returns:
            A list of response strings finishing with 'OK', or 
                ['ERROR', '<error_code>']
        
        Raises:
            AtException if no response was received.
            AtException if bad CRC response count exceeds retries

        """
        if current_thread() != self._thread:
            _log.warning('Call from external thread may crash or hang')
            loop = get_running_loop()
            set_event_loop(loop)
            await sleep(1)   #: add a slight delay to mitigate race condition
            while self._event.is_set():
                pass
            concurrentfuture = run_coroutine_threadsafe(
                self._command(at_command, timeout, retries), loop)
            asyncfuture = wrap_future(concurrentfuture)
            return await asyncfuture
        else:
            return await self._command(at_command, timeout, retries)

    async def _command(self,
                       at_command: str,
                       timeout: int,
                       retries: int) -> list:
        """Submits an AT command and returns the response asynchronously.
        
        Args:
            at_command: The AT command string
            timeout: The maximum time in seconds to await a response.
            retries: Optional number of additional attempts on failure.
        
        Returns:
            A list of response strings finishing with 'OK', or 
                ['ERROR', '<error_code>']
        
        Raises:
            AtException if no response was received.
            AtException if bad CRC response count exceeds retries

        """
        try:
            self._event.set()
            try:
                if self._verbose:
                    _log.debug('Opening serial port {}'.format(self.port))
                self._serial = AioSerial(port=self.port,
                                            baudrate=self.baudrate,
                                            loop=self.loop)
            except Exception as e:
                _log.error('Error connecting to aioserial: {}'.format(e))
            try:
                if self._verbose:
                    _log.debug('Checking unsolicited data'
                               f' prior to {at_command}')
                self._pending_command_time = time()
                unsolicited = await self._recv(timeout=0.25)
                if unsolicited:
                    _log.warning('Unsolicited data: {}'.format(unsolicited))
                    # raise AtUnsolicited('Unsolicited data: {}'.format(unsolicited))
            except AtTimeout:
                if self._verbose:
                    _log.debug('No unsolicited data found')
            tasks = [self._send(at_command),
                self._recv(timeout=timeout)]
            echo, response = await gather(*tasks)
            if echo in response:
                response.remove(echo)
            if len(response) > 0:
                self._retry_count = 0
                if response[0] == 'ERROR':
                    _log.debug('AT error detected - getting reason')
                    error_code = await self.command('ATS80?')
                    if error_code is not None:
                        response.append(error_code[0])
                    else:
                        _log.error('Failed to get error_code from S80')
                return response
            raise AtException('No response received for {}'.format(at_command))
        except AtCrcError:
            self._retry_count += 1
            if self._retry_count < retries:
                _log.error('CRC error retrying')
                return await self.command(
                    at_command, timeout=timeout, retries=retries)
            else:
                error_message = 'Too many failed CRC ({})'.format(
                    self._retry_count)
                self._retry_count = 0
                raise AtException(error_message)
        finally:
            if self._serial:
                if self._verbose:
                    _log.debug('Closing serial port {}'.format(self.port))
                self._serial.close()
                self._serial = None
            self._event.clear()
    
    async def initialize(self, crc: bool = False) -> bool:
        """Initializes the modem using ATZ and sets up CRC.

        Args:
            crc: desired initial CRC enabled if True

        Returns:
            True if successful
        
        Raises:
            AtException on errors other than CRC enabled

        """
        _log.debug('Initializing modem{}'.format(
            ' (CRC enabled)' if crc else ''))
        cmd = 'ATZ;E1;V1'
        cmd += ';%CRC=1' if crc else ''
        success = await self.command(cmd)
        if success[0] == 'ERROR':
            if int(success[1]) == 100:
                if crc and self.crc:
                    _log.debug('CRC already enabled')
                    return True
                else:
                    self.crc = True
                    await self.initialize(crc)
            else:
                return self._handle_at_error(cmd, success[1], return_value=False)
        self.crc = crc
        return True
    
    async def config_restore_nvm(self) -> bool:
        """Sends the ATZ command to restore from non-volatile memory.
        
        Returns:
            Boolean success.
        """
        _log.debug('Restoring non-volatile configuration')
        cmd = 'ATZ'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], return_value=False)
        return True

    async def config_restore_factory(self) -> bool:
        """Sends the AT&F command and returns True on success."""
        _log.debug('Restoring factory defaults')
        cmd = 'AT&F'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], return_value=False)
        return True
    
    async def config_report(self) -> Tuple[dict, dict]:
        """Sends the AT&V command to retrive S-register settings.
        
        Returns:
            A tuple with two dictionaries or both None if failed
            at_config with booleans crc, echo, quiet and verbose
            reg_config with S-register tags and integer values

        """
        _log.debug('Querying configuration')
        cmd = 'AT&V'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], (None, None))
        at_config = response[1]
        s_regs = response[2]
        echo, quiet, verbose, crc = at_config.split(' ')
        at_config = {
            "crc": bool(int(crc[4])),
            "echo": bool(int(echo[1])),
            "quiet": bool(int(quiet[1])),
            "verbose": bool(int(verbose[1])),
        }
        reg_config = {}
        for reg in s_regs.split(' '):
            name, value = reg.split(':')
            reg_config[name] = int(value)
        return (at_config, reg_config)

    async def config_save(self) -> bool:
        """Sends the AT&W command and returns True if successful."""
        _log.debug('Saving S-registers to non-volatile memory')
        cmd = 'AT&W'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], False)
        return True

    async def config_crc_enable(self, crc: bool) -> bool:
        """Enables or disables CRC error checking (for long serial cable).
        
        Args:
            crc: enable CRC if true
        """
        _log.debug('{} CRC'.format('Enabling' if crc else 'Disabling'))
        cmd = 'AT%CRC={}'.format(1 if crc else 0)
        response = await self.command(cmd)
        if response[0] == 'ERROR' and self.crc != crc:
            return self._handle_at_error(cmd, response[1], False)
        self.crc = crc
        return True
    
    async def device_mobile_id(self) -> str:
        """Returns the unique Mobile ID (Inmarsat serial number).
        
        Returns:
            MobileID string.
        
        Raises:
            AtException

        """
        _log.debug('Querying device Mobile ID')
        cmd = 'AT+GSN'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            self._handle_at_error(cmd, response[1])
        return response[0].replace('+GSN:', '').strip()

    async def device_version(self) -> Tuple[str, str, str]:
        """Returns the hardware, firmware and AT versions.
        
        Returns:
            Dict with hardware, firmware, at version.
        
        Raises:
            AtException

        """
        _log.debug('Querying device version info')
        cmd = 'AT+GMR'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            self._handle_at_error(cmd, response[1])
        versions = response[0].replace('+GMR:', '').strip()
        fw_ver, hw_ver, at_ver = versions.split(',')
        return {'hardware': hw_ver, 'firmware': fw_ver, 'at': at_ver}

    async def gnss_continuous_set(self,
                                  interval: int=0,
                                  doppler: bool=True) -> bool:
        """Sets the GNSS continous mode (0 = on-demand).
        
        Args:
            interval: Seconds between GNSS refresh.
            doppler: Often required for moving assets.
        
        Returns:
            True if successful setting.

        """
        _log.debug('Setting GNSS refresh to {} seconds'.format(interval))
        cmd = 'AT%TRK={}{}'.format(interval, ',{}'.format(1 if doppler else 0))
        if interval < 0 or interval > 30:
            raise ValueError('GNSS continuous interval must be in range 0..30')
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], False)
        return True

    async def gnss_nmea_get(self,
                            stale_secs: int = 1,
                            wait_secs: int = 35,
                            sentences: list = ['RMC', 'GSA', 'GGA', 'GSV']
                            ) -> Union[list, str]:
        """Returns a list of NMEA-formatted sentences from GNSS.

        Args:
            stale_secs: Maximum age of fix in seconds (1..600)
            wait_secs: Maximum time to wait for fix (1..600)
            sentences: Optional list of NMEA sentence types to get

        Returns:
            List of NMEA sentences

        Raises:
            ValueError if parameter out of range
            AtGnssTimeout if no response from GNSS
            AtException

        """
        _log.debug('Requesting GNSS fix information')
        NMEA_SUPPORTED = ['RMC', 'GGA', 'GSA', 'GSV']
        BUFFER_SECONDS = 5
        if (stale_secs not in range(1, 600+1) or
            wait_secs not in range(1, 600+1)):
            raise ValueError('stale_secs and wait_secs must be 1..600')
        sentence_list = ''
        for sentence in sentences:
            sentence = sentence.upper()
            if sentence in NMEA_SUPPORTED:
                if len(sentence_list) > 0:
                    sentence_list += ','
                sentence_list += '"{}"'.format(sentence)
            else:
                raise ValueError('Unsupported NMEA sentence: {}'
                                 .format(sentence))
        cmd = 'AT%GPS={},{},{}'.format(stale_secs, wait_secs, sentence_list)
        response = await self.command(cmd, timeout=wait_secs + BUFFER_SECONDS)
        if response[0] == 'ERROR':
            if int(response[1]) == 108:
                raise AtGnssTimeout('Timed out waiting for GNSS fix')
            else:
                return self._handle_at_error(cmd, response[1], None)
        if 'OK' in response:
            response.remove('OK')
        response[0] = response[0].replace('%GPS: ', '')
        return response

    async def location(self,
                       stale_secs: int = 1,
                       wait_secs: int = 35) -> Location:
        """Returns a location object.
        
        Args:
            stale_secs: the maximum fix age to accept
            wait_secs: the maximum time to wait for a new fix
        
        Returns:
            nmea.Location object
        
        Raises:
            AtGnssTimeout if no location data is available
        
        """
        _log.debug('Querying location')
        nmea_sentences = await self.gnss_nmea_get(stale_secs, wait_secs)
        return location_from_nmea(nmea_sentences)

    async def lowpower_mode_set(self, power_mode: int) -> bool:
        """Sets the modem power mode (for blockage recovery).

        Args:
            power_mode (int): The new power mode

        Returns:
            True if successful
        
        Raises:
            ValueError on invalid power_mode
        """
        if power_mode not in POWER_MODES:
            raise ValueError('Invalid power mode {}'.format(power_mode))
        _log.debug('Setting power mode {}'.format(
            POWER_MODES[power_mode]))
        cmd = 'ATS50={}'.format(power_mode)
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], False)
        return True

    async def lowpower_mode_get(self) -> int:
        """Gets the modem power mode.

        Returns:
            The integer value of the power mode
        
        Raises:
            AtException if an error was returned

        """
        _log.debug('Getting power mode')
        cmd = 'ATS50?'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], None)
        return int(response[0])

    async def lowpower_wakeup_set(self, wakeup_period: int) -> bool:
        """Sets the modem wakeup period.

        Args:
            wakeup_period (int): The new wakeup period

        Returns:
            True if successful
        
        Raises:
            ValueError on invalid wakeup_period

        """
        if wakeup_period not in WAKEUP_PERIODS:
            raise ValueError('Invalid wakeup period {}'.format(wakeup_period))
        _log.debug('Setting wakeup period {}'.format(
            WAKEUP_PERIODS[wakeup_period]))
        cmd = 'ATS51={}'.format(wakeup_period)
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], False)
        return True

    async def lowpower_wakeup_get(self) -> int:
        """Gets the modem wakeup period.

        Returns:
            The integer value of the wakeup period
        
        Raises:
            AtException if an error was returned

        """
        _log.debug('Getting wakeup period')
        cmd = 'ATS51?'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], None)
        return int(response[0])

    async def lowpower_notifications_enable(self) -> bool:
        """Configures low power satellite status and notification assertion.

        The following events trigger assertion of the notification output:
        - New Forward Message received
        - Return Message completed (success or failure)
        - Trace event update (satellite status change)

        Returns:
            True if successful
        """
        _log.debug('Enabling low power notifications')
        cmd = 'AT%EVMON=3.1;S88=1030'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], False)
        return True

    async def lowpower_notifications_check(self) -> list:
        """Returns a list of relevant events."""
        _log.debug('Querying low power notifications')
        relevant = []
        try:
            reason = await self.notification_check()
            if reason is not None:
                if reason['event_cached'] == True:
                    relevant.append('event_cached')
                if reason['message_mt_received'] == True:
                    relevant.append('message_mt_received')
                if reason['message_mo_complete'] == True:
                    relevant.append('message_mo_complete')
        except AtException:
            _log.warning('Notification check returned AT exception')
        finally:
            return relevant

    async def message_mo_send(self,
                              data: str,
                              data_format: int,
                              sin: int,
                              min: int = None,
                              name: str = None,
                              priority: int = 4) -> str:
        """Submits a mobile-originated message to send.
        
        Args:
            data: The data to be sent formatted as base64, hex or text according
                to `data_format`.
            data_format: 1: Text, 2: ASCII-Hex, 3: Base64 (MIME)
            name: (Optional) A unique name for the message, if none is provided
                a name based on unix timestamp will be assigned
            priority: 1: High .. 4: Low (default)
            sin: Service Identification Number (15..255) becomes the first byte
                of message payload
            min: (Optional) Message Identification Number (0..255) becomes the
                second byte of message payload if specified

        Returns:
            Name of the message if successful, or the error string
        """
        _log.debug('Submitting message named {}'.format(name))
        if name is None:
            # Use the 8 least-signficant numbers of unix timestamp as unique
            name = str(int(time()))[-8:]
            _log.debug('Assigned name {}'.format(name))
        elif len(name) > 8:
            name = name[0:8]   # risk duplicates create an ERROR resposne
            _log.warning('Truncated name to {}'.format(name))
        _min = '.{}'.format(min) if min is not None else ''
        if data_format == 1:
            data = '"{}"'.format(data)
        cmd = ('AT%MGRT="{}",{},{}{},{},{}'.format(name,
                                                    priority,
                                                    sin,
                                                    _min,
                                                    data_format,
                                                    data))
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], None)
        return name

    async def message_mo_state(self, name: str = None) -> list:
        """Returns the message state(s) requested.
        
        If no name filter is passed in, all available messages states
        are returned.  Returns False is the request failed.

        Args:
            name: The unique message name in the modem queue

        Returns:
            `list` of `dict` with `name`, `state`, `size` and `sent`

        Raises:
            AtException

        """
        _log.debug('Querying transmit message state{}'.format(
            ' ={}'.format(name) if name else 's'))
        cmd = 'AT%MGRS{}'.format('="{}"'.format(name) if name else '')
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], None)
        # %MGRS: "<name>",<msg_no>,<priority>,<sin>,<state>,<size>,<sent_bytes>
        if 'OK' in response:
            response.remove('OK')
        states = []
        for res in response:
            res = res.replace('%MGRS:', '').strip()
            if len(res) > 0:
                name, number, priority, sin, state, size, sent = res.split(',')
                del number
                del priority
                del sin
                states.append({
                    'name': name.replace('"', ''),
                    'state': int(state),
                    'size': int(size),
                    'bytes_sent': int(sent),
                    })
        return states
    
    @staticmethod
    def message_state_name(state: int):
        return MessageState(state).name

    async def message_mo_cancel(self, name: str) -> bool:
        """Cancels a mobile-originated message in the Tx ready state."""
        _log.debug('Cancelling message {}'.format(name))
        cmd = 'AT%MGRC="{}"'.format(name)
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], False)
        return True

    async def message_mo_clear(self) -> int:
        """Clears the modem transmit queue.
        
        Returns:
            Count of messages deleted
        
        Raises:
            AtException

        """
        _log.debug('Clearing transmit queue of return messages')
        cancelled_count = 0
        open_count = 0
        cmd = 'AT%MGRS'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1])
        if 'OK' in response:
            response.remove('OK')
        if '%MGRS:' in response:
            response.remove('%MGRS:')
        for message in response:
            if '%MGRS:' in message:
                message = message.replace('%MGRS:', '').strip()
            parts = message.split(',')
            status = int(parts[4])
            name = parts[0].replace('"', '')
            if status < 6:
                cancel_explicit = await self.message_mo_cancel(name)
                if not cancel_explicit:
                    open_count += 1
                else:
                    cancelled_count += 1
        if open_count > 0:
            _log.warning('{} messages still in transmit queue'.format(
                open_count))
        return cancelled_count

    async def message_mt_waiting(self) -> list:
        """Returns a list of received mobile-terminated message information.
        
        Returns:
            List of (name, number, priority, sin, state, length, received)
        
        Raises:
            AtException

        """
        _log.debug('Checking receive queue for forward messages')
        cmd = 'AT%MGFN'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1])
        if 'OK' in response:
            response.remove('OK')
        waiting = []
        #: %MGFN: name, number, priority, sin, state, length, bytes_received
        for res in response:
            msg = res.replace('%MGFN:', '').strip()
            if msg.startswith('"FM'):
                parts = msg.split(',')
                name, number, priority, sin, state, length, received = parts
                del number   #: unused
                waiting.append({'name': name.replace('"', ''),
                                'sin': int(sin),
                                'priority': int(priority),
                                'state': int(state),
                                'length': int(length),
                                'received': int(received)})
        return waiting

    @staticmethod
    def _message_mt_parse(mgfg_response: str,
                          data_format: int) -> dict:
        #:%MGFG:"<msgName>",<msgNum>,<priority>,<sin>,<state>,<length>,<data_format>,<data>
        parts = mgfg_response.replace('%MGFG:', '').strip().split(',')
        sys_msg_num, sys_msg_seq = parts[1].split('.')
        msg_sin = int(parts[3])
        data_str_no_sin = parts[7]
        if data_format == DataFormat.HEX:
            data = '{:02X}'.format(msg_sin) + data_str_no_sin
            databytes = bytes.fromhex(data)
        elif data_format == DataFormat.BASE64:
            databytes = bytes([msg_sin]) + b64decode(data_str_no_sin)
            data = b64encode(databytes).decode('ascii')
        elif data_format == DataFormat.TEXT:
            data_str_no_sin = data_str_no_sin[1:len(data_str_no_sin) - 1]
            data = '\\{:02x}'.format(msg_sin) + data_str_no_sin
            databytes = bytes([msg_sin])
            i = 0
            while i < len(data_str_no_sin):
                if data_str_no_sin[i] == '\\' and i < len(data_str_no_sin) - 1:
                    if data_str_no_sin[i + 1] in '0123456789ABCDEF':
                        databytes += bytes([int(data_str_no_sin[i+1:i+3], 16)])
                        i += 3
                else:
                    databytes += data_str_no_sin[i].encode('utf-8')
                    i += 1
        return {
            'name': parts[0].replace('"', ''),
            'system_message_number': int(sys_msg_num),
            'system_message_sequence': int(sys_msg_seq),
            'priority': int(parts[2]),
            'sin': msg_sin,
            'min': databytes[1],
            'state': int(parts[4]),
            'length': int(parts[5]),
            'data_format': data_format,
            'raw_payload': data,
            'bytes': databytes,
        }

    async def message_mt_get(self,
                             name: str,
                             data_format: int = DataFormat.BASE64,
                             verbose: bool = True) -> Union[dict, bytes]:
        """Returns the payload of a specified mobile-terminated message.
        
        Payload is presented as a string with encoding based on data_format. 

        Args:
            name: The unique name in the modem queue e.g. FM01.01
            data_format: text=1, hex=2, base64=3 (default)
            verbose: if True returns a dictionary, otherwise raw payload bytes

        Returns:
            The encoded data as a string
        
        Raises:
            AtException

        """
        _log.debug('Retrieving forward message {}'.format(name))
        cmd = 'AT%MGFG="{}",{}'.format(name, data_format)
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1])
        message = self._message_mt_parse(response[0], data_format=data_format)
        return message if verbose else message['bytes']

    async def message_mt_delete(self, name: str) -> bool:
        """Marks a Return message for deletion by the modem.
        
        Args:
            name: The unique mobile-terminated name in the queue

        Returns:
            True if the operation succeeded

        """
        _log.debug('Marking forward message {} for deletion'.format(name))
        cmd = 'AT%MGFM="{}"'.format(name)
        try:
            response = await self.command(cmd)
            if response[0] == 'ERROR':
                return self._handle_at_error(cmd, response[1], False)
            return True
        except:
            return False

    async def event_monitor_get(self) -> list:
        """Returns a list of monitored/cached events.
        As a list of <class.subclass> strings which includes an asterisk
        for each new event that can be retrieved.

        Returns:
            list of strings <class.subclass[*]> or None
        
        Raises:
            AtException

        """
        _log.debug('Querying monitored events')
        cmd = 'AT%EVMON'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1])
        events = response[0].replace('%EVMON: ', '').split(',')
        '''
        for i in range(len(events)):
            c, s = events[i].strip().split('.')
            if s[-1] == '*':
                s = s.replace('*', '')
                # TODO flag change for retrieval
            events[i] = (int(c), int(s))
        '''
        return [event for event in events if event != '']

    async def event_monitor_set(self, eventlist: list) -> bool:
        """Sets trace events to monitor.

        Args:
            eventlist: list of tuples (class, subclass)

        Returns:
            True if successfully set

        """
        _log.debug('Setting event monitors: {}'.format(eventlist))
        #: AT%EVMON{ = <c1.s1>[, <c2.s2> ..]}
        cmd = 'AT%EVMON='
        if eventlist is not None:
            for monitor in eventlist:
                if isinstance(monitor, tuple):
                    if len(cmd) > 9:
                        cmd += ','
                    cmd += '{}.{}'.format(monitor[0], monitor[1])
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], False)
        return True

    async def event_get(self,
                        event: tuple,
                        raw: bool = True) -> Union[str, dict]:
        """Gets the cached event by class/subclass.

        Args:
            event: tuple of (class, subclass)
            raw: Returns the raw text string if True
        
        Returns:
            String if raw=True, dictionary if raw=False
        
        Raises:
            AtException

        """
        _log.debug('Querying events: {}'.format(event))
        #: AT%EVNT=c,s
        #: res %EVNT: <dataCount>,<signedBitmask>,<MTID>,<timestamp>,
        # <class>,<subclass>,<priority>,<data0>,<data1>,..,<dataN>
        if not (isinstance(event, tuple) and len(event) == 2):
            raise AtException('event_get expects (class, subclass)')
        cmd = 'AT%EVNT={},{}'.format(event[0], event[1])
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1])
        eventdata = response[0].replace('%EVNT: ', '').split(',')
        event = {
            'data_count': int(eventdata[0]),
            'signed_bitmask': bin(int(eventdata[1]))[2:],
            'mobile_id': eventdata[2],
            'timestamp': eventdata[3],
            'class': eventdata[4],
            'subclass': eventdata[5],
            'priority': eventdata[6],
            'data': eventdata[7:]
        }
        bitmask = event['signed_bitmask']
        while len(bitmask) < event['data_count']:
            bitmask = '0' + bitmask
        i = 0
        for bit in reversed(bitmask):
            #: 32-bit signed conversion redundant since response is string
            if bit == '1':
                event['data'][i] = _to_signed32(int(event['data'][i]))
            else:
                event['data'][i] = int(event['data'][i])
            i += 1
        # TODO lookup class/subclass definitions
        return response[0] if raw else event

    async def notification_control_set(self, event_map: list) -> bool:
        """Sets the event notification bitmask.

        Args:
            event_map: list of tuples (event_name, bool)
        
        Returns:
            True if successful.
            
        """
        _log.debug('Setting event notifications: {}'.format(event_map))
        #: ATS88=bitmask
        notifications_changed = False
        old_notifications = await self.notification_control_get()
        if old_notifications is None:
            return False
        bitmask = list('0' * len(old_notifications))
        i = 0
        for event in event_map:
            if event[0] not in NOTIFICATION_BITMASK:
                raise ValueError('Invalid event {}'.format(event[0]))
            i = 0
            for key in reversed(old_notifications):
                bit = '1' if old_notifications[key] or bitmask[i] == '1' else '0'
                if key == event[0]:
                    notify = event[1]
                    if old_notifications[key] != notify:
                        bit = '1' if notify else '0'
                        notifications_changed = True
                        # self.notifications[key] = notify
                bitmask[i] = bit
                i += 1
        if notifications_changed:
            cmd = 'ATS88={}'.format(int('0b' + ''.join(bitmask), 2))
            response = await self.command(cmd)
            if response[0] == 'ERROR':
                return self._handle_at_error(cmd, response[1], False)
        return True
    
    async def notification_control_get(self) -> OrderedDict:
        """Returns the current notification configuration bitmask.
        
        Returns:
            OrderedDict
        
        Raises:
            AtException

        """
        _log.debug('Querying event notification controls')
        cmd =  'ATS88?'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1])
        return _notifications_dict(int(response[0]))

    async def notification_check(self) -> OrderedDict:
        """Returns the current active event notification bitmask (S89).
        
        The value of S89 register is cleared upon reading.

        Returns:
            OrderedDict
        
        Raises:
            AtException

        """
        _log.debug('Querying event notification triggers')
        cmd = 'ATS89?'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1])
        return _notifications_dict(int(response[0]))

    async def satellite_status(self) -> dict:
        """Returns the control state and C/No.
        
        Returns:
            Dictionary with state (int), snr (float), beamsearch (int),
                state_name (str), beamsearch_name (str), or None if error.

        Raises:
            AtException

        """
        _log.debug('Querying satellite status/SNR')
        cmd = 'ATS90=3 S91=1 S92=1 S116? S122? S123?'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1])
        if 'OK' in response:
            response.remove('OK')
        cn_0, ctrl_state, beamsearch_state = response
        cn_0 = int(cn_0) / 100.0
        ctrl_state = int(ctrl_state)
        beamsearch_state = int(beamsearch_state)
        return {
            'state': ctrl_state,
            'state_name': CONTROL_STATES[ctrl_state],
            'snr': cn_0,
            'beamsearch': beamsearch_state,
            'beamsearch_name': BeamSearchState(beamsearch_state).name,
        }

    @staticmethod
    def sat_status_name(ctrl_state: int) -> str:
        """Returns human-readable definition of a control state value.
        
        Raises:
            ValueError if ctrl_state is not found.
        """
        if ctrl_state not in CONTROL_STATES:
            raise ValueError('Control state {} not found'.format(ctrl_state))
        return CONTROL_STATES[ctrl_state]

    @staticmethod
    def sat_beamsearch_name(beamsearch_state: int) -> str:
        return BeamSearchState(beamsearch_state).name

    async def transmit_status(self) -> dict:
        """Returns the transmitter status.
        
        Returns:
            Transmit status (5 = OK)

        Raises:
            AtException if error returned by modem

        """
        _log.debug('Querying transmitter status')
        cmd = 'ATS54?'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1])
        status = int(response[0])
        return TransmitterStatus(status)

    async def shutdown(self) -> bool:
        """Tell the modem to prepare for power-down."""
        _log.debug('Requesting power down')
        cmd = 'AT%OFF'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1], False)
        return True

    async def time_utc(self) -> str:
        """Returns current UTC time of the modem in ISO format.
        
        Returns:
            UTC as ISO-formatted string
        
        Raises:
            AtException

        """
        _log.debug('Requesting UTC network time')
        cmd = 'AT%UTC'
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1])
        return response[0].replace('%UTC: ', '').replace(' ', 'T') + 'Z'

    async def s_register_get(self, register: int) -> Union[int, None]:
        """Returns the value of the S-register requested.

        Args:
            register: The S-register number

        Returns:
            integer value of register
        
        Raises:
            AtException

        """
        _log.debug('Querying register value S{}'.format(register))
        cmd = 'ATS{}?'.format(register)
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[1])
        return int(response[0])

    async def s_register_get_all(self) -> list:
        """Returns a list of S-register definitions.
        R=read-only, S=signed, V=volatile
        
        Returns:
            tuple(register, RSV, current, default, minimum, maximum)
        
        Raises:
            AtException

        """
        _log.debug('Querying S-register values')
        cmd = 'AT%SREG'
        #: Sreg, RSV, CurrentVal, DefaultVal, MinimumVal, MaximumVal
        response = await self.command(cmd)
        if response[0] == 'ERROR':
            return self._handle_at_error(cmd, response[0])
        if 'OK' in response:
            response.remove('OK')
        reg_defs = response[2:]
        registers = []
        for row in reg_defs:
            reg_def = row.split(' ')
            reg_def = tuple(filter(None, reg_def))
            registers.append(reg_def)
        return registers



"""Base classes for AT command errors.
"""

class AtException(Exception):
    """Base class for AT command exceptions."""


class AtTimeout(AtException):
    """Indicates a timeout waiting for response."""


class AtGnssTimeout(AtException):
    """Indicates a timeout waiting for a GNSS fix."""


class AtCrcError(AtException):
    """Indicates a detected CRC mismatch on a response."""


class AtCrcConfigError(AtException):
    """Indicates a CRC response was received when none expected or vice versa.
    """


class AtUnsolicited(AtException):
    """Indicates unsolicited data was received from the modem."""



"""A threaded IDP modem client with abstracted properties."""
import logging
import os
import queue
from base64 import b64decode, b64encode
from datetime import datetime, timezone
from math import ceil
from time import time

from idpmodem.aterror import AtCrcError, AtException, AtGnssTimeout, AtTimeout
from idpmodem.constants import (EVENT_TRACES, AtErrorCode, BeamSearchState,
                                DataFormat, EventNotification, GeoBeam,
                                GnssMode, MessagePriority, MessageState,
                                PowerMode, SatlliteControlState,
                                SignalLevelRegional, SignalQuality,
                                TransmitterStatus, WakeupPeriod)
from idpmodem.helpers import printable_crlf
from idpmodem.location import Location, location_from_nmea
from idpmodem.s_registers import SRegisters
from idpmodem.threaded.atcommand import AtProtocol, ByteReaderThread, Serial

GNSS_STALE_SECS = int(os.getenv('GNSS_STALE_SECS', 1))
GNSS_WAIT_SECS = int(os.getenv('GNSS_WAIT_SECS', 35))
SAT_STATUS_HOLDOFF = 5
MODEM_REBOOT_HOLDOFF = os.getenv('MODEM_REBOOT_HOLDOFF')
VERBOSE_DEBUG = str(os.getenv('VERBOSE_DEBUG', False)).lower() == 'true'

_log = logging.getLogger(__name__)


class ModemBusy(Exception):
    """Indicates the modem is busy processing a prior command."""


class AtConfiguration:
    """Configuration settings of the modem.
    
    Attributes:
        crc (bool): Using cyclic redundancy check for all transactions.
        echo (bool): Echoing back commands
        quiet (bool): Limiting responses
        verbose (bool): Using text-based responses
        
    """
    def __init__(self) -> None:
        self.crc: bool = False
        self.echo: bool = True
        self.quiet: bool = False
        self.verbose: bool = True


class IdpModem:
    """Abstracts AT commands to relevant functions for an IDP modem.
    
    Attributes:
        connected (bool): Indicates if connected to a modem on serial.
        baudrate (int): The baudrate of the modem.
        crc (bool): Indicates if CRC error checking is enabled.
        mobile_id (str): The unique modem ID.
        versions (dict): The versions reported by the modem.
        manufacturer (str): The modem manufacturer.
        model (str): The modem model.
        power_mode (IntEnum):
        wakeup_period (IntEnum):
        temperature (float):
        gnss_refresh_interval (int):
        location (object):
        control_state (IntEnum):
        beam_search_state (IntEnum):
        network_status (IntEnum):
        registered (bool):
        
    """
    
    SERIAL_KWARGS = ['baudrate', 'timeout', 'write_timeout']
    BAUD_RATES = [1200, 2400, 4800, 9600, 19200]
    PROTOCOL_KWARGS = ['event_callback', 'at_timeout']
    OTHER_KWARGS = ['error_detail', 'stale_secs', 'wait_secs']
    
    def __init__(self, serial_port: str, **kwargs):
        self.serial_kwargs = {
            'port': serial_port,
            'baudrate': int(kwargs.pop('baudrate', 9600)),
        }
        self.protocol_kwargs = {}
        self.error_detail = bool(kwargs.pop('error_detail', True))
        for kwarg in kwargs:
            if kwarg in self.SERIAL_KWARGS:
                self.serial_kwargs[kwarg] = kwargs[kwarg]
            elif kwarg in self.PROTOCOL_KWARGS:
                self.protocol_kwargs[kwarg] = kwargs[kwarg]
        self.serial_port = None
        self.main_thread = None
        self.transport = None
        self.protocol: AtProtocol = None
        try:
            self._reboot_holdoff = int(MODEM_REBOOT_HOLDOFF)
        except:
            self._reboot_holdoff = None
        self.commands = queue.Queue(1)
        self._at_config = AtConfiguration()
        self._mobile_id: str = None
        self._versions: dict = None
        self._manufacturer: str = None
        self._model: str = None
        self._power_mode: int = None
        self._wakeup_period: int = None
        self._ctrl_state: int = None
        self._snr: float = None
        self._geo_beam_id: int = None
        self._beamsearch_state: int = None
        self._loc_query: dict = {
            'stale_secs': int(kwargs.pop('stale_secs', GNSS_STALE_SECS)),
            'wait_secs': int(kwargs.pop('wait_secs', GNSS_WAIT_SECS)),
        }
        self._holdoffs: dict = {}   # used to ignore frequent repeat commands
        self._statistics: dict = {}
        self.s_registers = SRegisters()
    
    def connect(self):
        """Connects to a modem using a serial and protocol instance."""
        self.serial_port = Serial(**self.serial_kwargs)
        self.main_thread = ByteReaderThread(self.serial_port,
                                            AtProtocol,
                                            **self.protocol_kwargs)
        self.main_thread.start()
        self.transport, self.protocol = self.main_thread.connect()
        assert isinstance(self.protocol, AtProtocol)
        self.serial_port.reset_input_buffer()
        self.serial_port.reset_output_buffer()
        if self._reboot_holdoff is not None:
            self.protocol.event_callback = self._unsolicited

    def disconnect(self):
        """Disconnects from the modem."""
        with self.commands.mutex:
            self.commands.queue.clear()
        if self.main_thread:
            self.main_thread.close()
        if self.serial_port:
            self.serial_port.close()
        self.transport = None
        self.protocol = None
    
    @property
    def connected(self) -> bool:
        """Indicates if the modem is connected.
        
        Attempts to send a basic `AT` command and check for any response.

        """
        if self.transport is None or self.protocol is None:
            return False
        try:
            res = self.atcommand('AT')
            if res is not None:
                return True
        except AtCrcError:
            return True
        except AtTimeout:
            pass
        return False

    @property
    def baudrate(self) -> 'int|None':
        """The baud rate of the serial connecton."""
        return self.serial_port.baudrate if self.serial_port else None
    
    @baudrate.setter
    def baudrate(self, value: int):
        if not self.connected:
            raise ConnectionError('Modem is not connected')
        if value not in self.BAUD_RATES:
            raise ValueError(f'Baud rate must be one of {self.BAUD_RATES}')
        response = self.atcommand(f'AT+IPR={value}')
        if response and response[0] != 'ERROR':
            self.serial_port.baudrate = value

    @property
    def crc(self) -> 'bool|None':
        """Indicates if CRC error checking is enabled on the modem."""
        return self.protocol.crc if self.protocol is not None else None

    def _unsolicited(self, data: str) -> None:
        if self.commands.full():
            raise ModemBusy('Unsolicited data received during AT command'
                            f'processing: {printable_crlf(data)}')
        if self._reboot_holdoff:
            BOOT_INDICATORS = ['boot loader', 'Copyright (c)', '*** Reset',
                'starting appl firmware', '.....']
            if any(indicator in data for indicator in BOOT_INDICATORS):
                _log.warning('Reboot indicator found - holding off commands'
                            f' {self._reboot_holdoff}s')
                self._holdoffs['reboot'] = int(time())

    def atcommand(self,
                  command: str,
                  filter: 'list[str]' = [],
                  timeout: int = 5,
                  await_previous: bool = True,
                  ) -> 'list[str]':
        """Sends an AT command to the modem and returns the response.
        
        Args:
            command: The AT command
            filter: (optional) list of sub/strings to remove from response.
            timeout: Number of seconds to wait for a reply
                (not including messages queued by other threads)
            await_previous: If True, this will block if a prior command was
                submitted by another thread
        
        Returns:
            list of filtered and stripped response(s) to the command(s)
        
        Raises:
            ModemBusy if await_previous is False and a prior command is queued.
            AtException if an error occurred that is unrecognized.

        """
        if not self.transport or not self.protocol:
            raise ConnectionError('No serial or protocol instance.')
        while self.commands.full():
            if not await_previous:
                raise ModemBusy
            pass
        if 'reboot' in self._holdoffs and isinstance(self._reboot_holdoff, int):
            while int(time()) - self._holdoffs['reboot'] < self._reboot_holdoff:
                pass
        self.commands.put(command)
        try:
            res: list = self.protocol.command(command,
                                              filter=filter,
                                              timeout=timeout)
            if VERBOSE_DEBUG:
                _log.debug(f'Response: {res}')
            if self.error_detail and res and res[0] == 'ERROR':
                _log.debug(f'Querying error code response to {command}')
                err_res = self.protocol.command('ATS80?')
                if not err_res or err_res[0] == 'ERROR':
                    raise AtException('Unhandled error getting last error code'
                                      f' ({err_res})')
                last_err_code = err_res[0]
                detail = 'UNDEFINED'
                if AtErrorCode.is_valid(int(last_err_code)):
                    detail = AtErrorCode(int(last_err_code)).name
                res.append(f'{detail} ({last_err_code})')
                _log.warning(f'AT error: {detail} for command {command}')
            return res
        except AtException as err:
            _log.error(f'{err} on command {command}')
            raise err
        finally:
            self.commands.get()
            self.commands.task_done()
    
    def _handle_at_error(self, response: 'list[str]') -> None:
        err = response[1] if self.error_detail else response[0]
        _log.error(f'AT Error: {err}')
        raise AtException(err)

    def config_init(self, crc: bool = False) -> bool:
        """Initializes modem communications with Echo, Verbose. CRC optional."""
        _log.debug(f'Initializing modem Echo|Verbose{"|CRC" if crc else ""}'
                   f' (CRC={self.protocol.crc})')
        command = f'ATZ;E1;V1;Q0;%CRC={1 if crc else 0}'
        res_attempt_1 = self.atcommand(command)
        if res_attempt_1[0] != 'OK':
            # case 1: crc True but previously set; should now be T in factory
            # case 2: crc False but previously set; should now be F in factory
            if len(res_attempt_1) > 1:
                at_error = res_attempt_1[1]
                if ('INVALID_CRC' not in at_error and
                    'UNKNOWN_COMMAND' not in at_error):
                    _log.warning(f'Unexpected AT error {at_error}')
            _log.debug(f'CRC mismatch, re-attempting (CRC={self.protocol.crc})')
            res_attempt_2 = self.atcommand(command)
            if res_attempt_2[0] != 'OK':
                _log.error('Unable to initialize modem after second attempt')
                if len(res_attempt_2) > 1:
                    _log.error(f'AT error: {res_attempt_2[1]}')
                return False
        # self.protocol.crc = crc   #: redundant should be set by attempt
        self._at_config.crc = crc
        _log.debug('Initialization success')
        return True

    def config_restore_nvm(self) -> bool:
        """Sends ATZ to restore config from non-volatile memory."""
        _log.debug('Restoring modem stored configuration')
        response = self.atcommand('ATZ')
        if response[0] == 'ERROR':
            return False
        return True

    def config_restore_factory(self) -> bool:
        """Sends AT&F to restore factory default and returns True on success."""
        _log.debug('Restoring modem factory defaults')
        response = self.atcommand('AT&F')
        if response[0] == 'ERROR':
            return False
        return True
    
    def config_report(self) -> 'tuple[dict, dict]':
        """Sends the AT&V command to retrieve S-register settings.
        
        Returns:
            A tuple with two dictionaries (empty if failed) with:
            at_config with booleans crc, echo, quiet and verbose
            reg_config with S-register tags and integer values
        
        Raises:
            AtException if an error was returned.

        """
        _log.debug('Retrieving modem verbose configuration')
        response = self.atcommand('AT&V')
        if response[0] == 'ERROR':
            self._handle_at_error(response)
        at_config = response[1]
        s_regs = response[2]
        echo, quiet, verbose, crc = at_config.split(' ')
        self._at_config.crc = bool(int(crc[4]))
        self._at_config.echo = bool(int(echo[1]))
        self._at_config.quiet = bool(int(quiet[1]))
        self._at_config.verbose = bool(int(verbose[1]))
        reg_config = {}
        for reg in s_regs.split(' '):
            name, value = reg.split(':')
            reg_config[name] = int(value)
        return (at_config, reg_config)

    def config_volatile_report(self) -> 'dict|None':
        """Gets key S-register settings.
        
        GNSS Mode (S39), GNSS fix timeout (S41), GNSS Continuous (S55),
        GNSS Jamming Status (S56), GNSS Jamming Indicator (S57), 
        Low power Wakeup Period (S51)

        Returns:
            Dictionary of S-register values, or None if failed
            
        """
        register_list = [
            'S39',   #: GNSS Mode
            'S41',   #: GNSS Fix Timeout
            'S51',   #: Wakeup Interval
            'S55',   #: GNSS Continuous
            'S56',   #: GNSS Jamming Status
            'S57',   #: GNSS Jamming Indicator
        ]
        _log.debug(f'Querying volatile S-register set: {register_list}')
        command = 'AT'
        for reg in register_list:
            command += f'{reg if command == "AT" else " " + reg}?'
        response = self.atcommand(command)
        if response[0] == 'ERROR':
            return None
        #: else
        response.remove('OK')
        volatile_regs = {}
        for r in range(len(response)):
            volatile_regs[register_list[r]] = int(response[r])
        return volatile_regs

    def config_nvm_save(self) -> bool:
        """Sends the AT&W command and returns True if successful."""
        _log.debug('Saving modem configuration to non-volatile memory')
        response = self.atcommand('AT&W')
        return response[0] == 'OK'

    def crc_enable(self, enable: bool = True) -> bool:
        """Sends the AT%CRC command and returns success flag.
        
        Args:
            enable: turn on CRC if True else turn off

        Returns:
            True if the operation succeeded else False

        """
        _log.debug(f'{"en" if enable else "dis"}abling modem CRC')
        command = f'AT%CRC={1 if enable else 0}'
        response = self.atcommand(command)
        if response[0] == 'ERROR':
            return False
        self.protocol.crc = enable
        self._at_config.crc = enable
        return True

    @property
    def mobile_id(self) -> 'str|None':
        """The unique Mobile ID (Inmarsat serial number)."""
        if self._mobile_id is None:
            if VERBOSE_DEBUG:
                _log.debug('Querying Mobile ID')
            response = self.atcommand('AT+GSN', filter=['+GSN:'])
            if response[0] != 'ERROR':
                self._mobile_id = response[0]
        return self._mobile_id

    @property
    def versions(self) -> 'dict|None':
        """The hardware, firmware and AT versions."""
        if not self._versions:
            if VERBOSE_DEBUG:
                _log.debug('Querying modem versions')
            response = self.atcommand('AT+GMR', filter=['+GMR:'])
            if response[0] != 'ERROR':
                self._versions = {}
                versions = response[0].split(',')
                if len(versions) == 3:
                    self._versions['firmware'] = versions[0]
                    self._versions['hardware'] = versions[1]
                    self._versions['at'] = versions[2]
                else:
                    for i, v in enumerate(versions):
                        self._versions[i] = v
        return self._versions

    @property
    def manufacturer(self) -> str:
        """The modem manufacturer reported by `ATI0`."""
        if not self._manufacturer:
            if VERBOSE_DEBUG:
                _log.debug('Querying modem manufacturer')
            response = self.atcommand('ATI0')
            if response[0] == 'ERROR':
                self._handle_at_error(response)
            self._manufacturer = response[0]
        return self._manufacturer

    @property
    def model(self) -> str:
        """The modem model reported by `ATI4`."""
        if not self._model:
            if VERBOSE_DEBUG:
                _log.debug('Querying modem model')
            response = self.atcommand('ATI4')
            if response[0] == 'ERROR':
                self._handle_at_error(response)
            self._model = response[0]
        return self._model

    @property
    def power_mode(self) -> 'PowerMode|None':
        """The modem power mode setting (enumerated) in `S50`."""
        if self._power_mode is None:
            if VERBOSE_DEBUG:
                _log.debug('Querying modem power mode')
            response = self.atcommand('ATS50?')
            if response[0] != 'ERROR':
                self._power_mode = PowerMode(int(response[0]))
        return self._power_mode
    
    @power_mode.setter
    def power_mode(self, value: 'str|int|PowerMode'):
        if isinstance(value, str):
            if value not in PowerMode.__members__:
                raise ValueError(f'Invalid PowerMode {value}')
            value = PowerMode[value].value
        if not PowerMode.is_valid(value):
            raise ValueError(f'Invalid PowerMode {value}')
        if VERBOSE_DEBUG:
            _log.debug(f'Setting modem power mode {PowerMode(value)}')
        response = self.atcommand(f'ATS50={value}')
        if response[0] == 'OK':
            self._power_mode = PowerMode(value)
    
    @property
    def wakeup_period(self) -> 'WakeupPeriod|None':
        """The modem wakeup period setting (enumerated) in `S51`."""
        if self._wakeup_period is None:
            if VERBOSE_DEBUG:
                _log.debug('Querying modem wakeup period')
            response = self.atcommand('ATS51?')
            if response[0] != 'ERROR':
                self._wakeup_period = WakeupPeriod(int(response[0]))
        return self._wakeup_period

    @wakeup_period.setter
    def wakeup_period(self, value: 'str|int|WakeupPeriod'):
        if isinstance(value, str):
            if value not in WakeupPeriod.__members__:
                raise ValueError(f'Invalid WakeupPeriod {value}')
            value = WakeupPeriod[value].value
        if not WakeupPeriod.is_valid(value):
            raise ValueError(f'Invalid WakeupPeriod {value}')
        if VERBOSE_DEBUG:
            _log.debug(f'Setting modem power mode {WakeupPeriod(value)}')
        response = self.atcommand(f'ATS51={value}')
        if response[0] == 'OK':
            self._wakeup_period = WakeupPeriod(value)
    
    @property
    def temperature(self) -> int:
        """Temperature in degrees Celsius (`S85`)."""
        if VERBOSE_DEBUG:
            _log.debug('Querying modem temperature')
        response = self.atcommand('ATS85?')
        if response[0] != 'ERROR':
            return int(float(response[0]) / 10)

    @property
    def gnss_refresh_interval(self) -> int:
        """GNSS refresh interval in seconds (`S55`)."""
        if VERBOSE_DEBUG:
            _log.debug('Querying GNSS refresh interval')
        response = self.atcommand(f'ATS55?')
        if response[0] != 'ERROR':
            return int(response[0])

    @gnss_refresh_interval.setter
    def gnss_refresh_interval(self, value: int):
        self.gnss_continuous_set(value)

    def gnss_continuous_set(self,
                            interval: int = 0,
                            doppler: bool = True,
                            ) -> bool:
        """Sets the GNSS continous mode (0 = on-demand).
        
        Args:
            interval: Seconds between GNSS refresh.
            doppler: Often required for moving assets.
        
        Returns:
            True if successful setting.
        """
        if interval < 0 or interval > 30:
            raise ValueError('GNSS continuous interval must be in range 0..30')
        _log.debug(f'Configuring GNSS continuous mode {interval} seconds')
        response = self.atcommand(f'AT%TRK={interval}{",1" if doppler else ""}')
        if response[0] == 'ERROR':
            return False
        return True

    def gnss_nmea_get(self,
                      stale_secs: int = GNSS_STALE_SECS,
                      wait_secs: int = GNSS_WAIT_SECS,
                      nmea: 'list[str]' = ['RMC', 'GSA', 'GGA', 'GSV'],
                      ) -> list:
        """Gets a list of NMEA-formatted sentences from GNSS.

        Args:
            stale_secs: Maximum age of fix in seconds (1..600)
            wait_secs: Maximum time to wait for fix (1..600)

        Returns:
            List of NMEA sentences

        Raises:
            ValueError if parameter out of range
            AtGnssTimeout if the fix timed out
            AtException if any other error code was returned

        """
        NMEA_SUPPORTED = ['RMC', 'GGA', 'GSA', 'GSV']
        BUFFER_SECONDS = 5
        if (stale_secs not in range(1, 600+1) or
            wait_secs not in range(1, 600+1)):
            raise ValueError('stale_secs and wait_secs must be 1..600')
        sentences = ''
        for sentence in nmea:
            sentence = sentence.upper()
            if sentence not in NMEA_SUPPORTED:
                raise ValueError(f'Unsupported NMEA sentence: {sentence}')
            if len(sentences) > 0:
                sentences += ','
            sentences += f'"{sentence}"'
        timeout = wait_secs + BUFFER_SECONDS
        request_time = time()
        _log.debug(f'Querying GNSS NMEA sentences {sentences}')
        response = self.atcommand(f'AT%GPS={stale_secs}'
                                        f',{wait_secs},{sentences}',
                                        timeout=timeout,
                                        filter=['%GPS:'])
        if response[0] == 'ERROR':
            if self.error_detail:
                if 'TIMEOUT' in response[1]:
                    raise AtGnssTimeout(response[1])
            self._handle_at_error(response)
        response.remove('OK')
        time_to_fix = round(time() - request_time, 3)
        if 'gnss_ttf' not in self._statistics:
            self._statistics['gnss_ttf'] = time_to_fix
        else:
            old_ttf = self._statistics['gnss_ttf']
            avg_ttf = round((time_to_fix + old_ttf) / 2, 3)
            self._statistics['gnss_ttf'] = avg_ttf
        return response

    @property
    def location(self) -> 'Location|None':
        """The modem location derived from NMEA data."""
        try:
            nmea_sentences = self.gnss_nmea_get(self._loc_query['stale_secs'],
                                                self._loc_query['wait_secs'])
            return location_from_nmea(nmea_sentences)
        except:
            return None

    @property
    def gnss_jamming(self) -> bool:
        """The GNSS jamming detection status (`S56`)."""
        if VERBOSE_DEBUG:
            _log.debug('Querying GNSS jamming indicator')
        response = self.atcommand('ATS56?')
        if response[0] != 'ERROR':
            return ((int(response[0]) & 0b100) >> 2 == 1) 

    @property
    def gnss_mode(self) -> GnssMode:
        """The GNSS operating mode setting (`S39`)."""
        if VERBOSE_DEBUG:
            _log.debug('Querying GNSS mode')
        response = self.atcommand('ATS39?')
        if response[0] != 'ERROR':
            return GnssMode(int(response[0]))

    @gnss_mode.setter
    def gnss_mode(self, mode: GnssMode):
        if VERBOSE_DEBUG:
            _log.debug(f'Setting GNSS mode {mode}')
        response = self.atcommand(f'ATS39={mode.value}')
        if response[0] == 'ERROR':
            self._handle_at_error(response)

    def message_mo_send(self,
                        data: 'bytes|bytearray|str',
                        data_format: int = DataFormat.BASE64,
                        name: str = None,
                        priority: int = MessagePriority.LOW,
                        sin: int = None,
                        min: int = None,
                        timeout: int = None,
                        ) -> str:
        """Submits a mobile-originated message to send.

        When submitting raw bytes, the first byte will be used as SIN. The
        first byte must not be in the reserved range (0..15).
        When submitting a string, the `sin` field is expected to be set and
        the data field will be appended to the `sin` byte and optionally the
        `min` byte if specified.
        
        Args:
            data: The data raw bytes or UTF-8 Text, Hexadecimal or Base64 string
            data_format: 1=text, 2=hexadecimal, 3=base64 (default)
            name: Optional unique name up to 8 characters long. If none is
                specified, use the 8 least-significant digits of unix timestamp.
            priority: 1=high, 4=low (default)
            sin: Optional first byte of payload used for codec, required if data
                is string type.
            min: Optional second byte of payload used for codec
            timeout: Optional timeout. If not provided will be calculated from
                the baudrate for the maximum message size, *3

        Returns:
            Name of the message if successful, or the error string.
        
        Raises:
            AtException if an error was returned by the modem.

        """
        name = str(int(time()))[-8:] if not name else name[0:8]
        if isinstance(data, bytes) or isinstance(data, bytearray):
            sin = data[0]
            data = b64encode(data[1:]).decode('utf-8')
            data_format = DataFormat.BASE64
        elif not isinstance(data, str):
            raise ValueError('Invalid data must be bytes, bytearray or string')
        if not isinstance(sin, int) or sin not in range(16, 256):
            raise ValueError('Invalid SIN must be 16..255')
        if isinstance(min, int) and min not in range(0, 256):
            raise ValueError('Invalid MIN must be 0..255')
        min = f'.{min}' if min is not None else ''
        data = f'"{data}"' if data_format == DataFormat.TEXT else data
        _log.debug(f'Submitting MO message with name {name}')
        command = f'AT%MGRT="{name}",{priority},{sin}{min},{data_format},{data}'
        max_timeout = timeout or ceil(6400 / (self.baudrate / 8)) * 3
        response = self.atcommand(command, timeout=max_timeout)
        if response[0] == 'ERROR':
            self._handle_at_error(response)
        return name

    def message_mo_state(self, name: str = None) -> 'list[dict]':
        """Gets the message state(s) requested.
        
        If no name filter is passed in, all available messages states
        are returned.  Returns False is the request failed.

        Args:
            name: The unique message name in the modem queue. If name is
                None, all available message states in transmit queue will be
                returned.

        Returns:
            List of metadata for each message in transmit queue including:
            - `name` (str) The ID in the modem transmit queue
            - `state` (int) The state of the message
            - `state_name` (str) The `MessageState.name`
            - `size` (int) in bytes
            - `sent` (int) in bytes for large message progress

        """
        states = []
        name = f'="{name}"' if name is not None else ''
        _log.debug(f'Querying MO message states {name if name else ""}')
        response = self.atcommand(f'AT%MGRS{name}', filter=['%MGRS:'])
        # %MGRS: "<name>",<msg_no>,<priority>,<sin>,<state>,<size>,<sent_bytes>
        if response[0] != 'ERROR':
            response.remove('OK')
            for msg in response:
                if ',' in msg:
                    detail = msg.split(',')
                    states.append({
                        'name': detail[0].replace('"', ''),
                        'state': int(detail[4]),
                        'state_name': MessageState(int(detail[4])).name,
                        'size': int(detail[5]),
                        'sent': int(detail[6]),
                    })
        return states
    
    def message_mo_cancel(self, name: str) -> bool:
        """Cancels a mobile-originated message in the Tx ready state."""
        _log.debug(f'Attempting to cancel message {name}')
        response = self.atcommand(f'AT%MGRC="{name}"')
        if response[0] == 'ERROR':
            return False
        return True

    def message_mo_clear(self) -> int:
        """Clears the modem transmit queue and returns the count cancelled.
        
        Returns:
            Count of messages deleted, or -1 in case of error

        """
        list_response = self.atcommand('AT%MGRL', filter=['%MGRL:'])
        if list_response[0] == 'ERROR':
            return -1
        message_count = len(list_response)
        for msg in list_response:
            _log.debug(f'Attempting to delete MO message {msg}')
            del_response = self.atcommand(f'AT%MGRD={msg}C')
            if del_response[0] == 'ERROR':
                _log.error(f'Error clearing messages from transmit queue')
                return -1
        return message_count

    def message_mt_waiting(self) -> 'list[dict]':
        """Gets a list of received mobile-terminated message information.
        
        Returns:
            List of message metadata in the receive queue including:
            - `name` (str)
            - `sin` (int) first byte of payload
            - `priority` (int)
            - `state` (int) The state number
            - `state_name` (str) The `MessageState.name`
            - `size` (int) in bytes
            - `received` (int) in bytes for large message progress

        """
        waiting = []
        _log.debug('Querying for waiting MT messages')
        response = self.atcommand('AT%MGFN', filter=['%MGFN:'])
        #: %MGFN: "name",number,priority,sin,state,length,bytes_received
        if response[0] != 'ERROR':
            response.remove('OK')
            for msg in response:
                if (',' in msg):
                    detail = msg.split(',')
                    waiting.append({
                        'name': detail[0].replace('"', ''),
                        'sin': int(detail[3]),
                        'priority': int(detail[2]),
                        'state': int(detail[4]),
                        'state_name': MessageState(int(detail[4])).name,
                        'size': int(detail[5]),
                        'received': int(detail[6])
                        })
        return waiting

    def message_mt_get(self,
                       name: str,
                       data_format: int = DataFormat.BASE64,
                       meta: bool = False,
                       timeout: int = None,
                       ) -> 'bytes|dict':
        """Gets the payload of a specified mobile-terminated message.
        
        Payload is presented as a string with encoding based on data_format. 

        Args:
            name: The unique name in the modem queue e.g. FM01.01
            data_format: text=1, hex=2, base64=3 (default)
            meta: If False returns raw bytes, else returns formatted data
                with metadata.
            timeout: Optional timeout. If not specified, will be calculated
                based on the baudrate for the maximum message size, *3

        Returns:
            The raw data bytes if meta is False, or a dictionary with:
            - `name` (str) The name assigned by the modem
            - `system_message_number` (int) System-assigned number
            - `system_message_sequence` (int) System-assigned number
            - `sin` (int) First byte of payload
            - `priority` (int)
            - `state` (int) The message state number
            - `state_name` (str) The `MessageState.name`
            - `size` (int) in bytes
            - `data_format` (int) 1=text, 2=hex, 3=base64
            - `data` (str) presented based on data_format

        """
        if not meta and data_format != DataFormat.BASE64:
            data_format = DataFormat.BASE64
        max_timeout = timeout or ceil(10000 / (self.baudrate / 8)) * 3
        _log.debug(f'Retrieving waiting MT message {name}')
        response = self.atcommand(f'AT%MGFG="{name}",{data_format}',
                                  filter=['%MGFG:'],
                                  timeout=max_timeout)
        if response[0] == 'ERROR':
            _log.error(f'Error retrieving message {name}')
            self._handle_at_error(response)
        #: name, number, priority, sin, state, length, data_format, data
        try:
            detail = response[0].split(',')
            sys_msg_num, sys_msg_seq = detail[1].split('.')
            msg_sin = int(detail[3])
            data_str_no_sin = detail[7]
            if data_format == DataFormat.HEX:
                data = hex(msg_sin) + data_str_no_sin.lower()
            elif data_format == DataFormat.BASE64:
                # add SIN byte to base64 blob
                databytes = bytes([msg_sin]) + b64decode(data_str_no_sin)
                if not meta:
                    return databytes
                data = b64encode(databytes).decode('ascii')
            elif data_format == DataFormat.TEXT:
                data = f'\\{msg_sin:02x}' + data_str_no_sin
            return {
                'name': detail[0].replace('"', ''),
                'system_message_number': int(sys_msg_num),
                'system_message_sequence': int(sys_msg_seq),
                'priority': int(detail[2]),
                'sin': msg_sin,
                'state': int(detail[4]),
                'state_name': MessageState(int(detail[4])).name,
                'size': int(detail[5]),
                'data_format': data_format,
                'data': data
            }
        except Exception as err:
            _log.exception(err)

    def message_mt_delete(self, name: str) -> bool:
        """Marks a Return message for deletion by the modem.
        
        Args:
            name: The unique mobile-terminated name in the queue

        Returns:
            True if the operation succeeded

        """
        _log.debug(f'Attempting to delete MT message {name}')
        response = self.atcommand(f'AT%MGFM="{name}"')
        if response[0] == 'ERROR':
            err = f' ({response[1]})' if self.error_detail else ''
            _log.error(f'Error deleting message {name}{err}')
        return response[0] == 'OK'

    @property
    def transmitter_status(self) -> TransmitterStatus:
        """The transmitter status reported by `S54`"""
        _log.debug('Querying transmitter status')
        response = self.atcommand('ATS54?')
        if response[0] == 'ERROR':
            self._handle_at_error(response)
        return TransmitterStatus(int(response[0]))

    def _trace_detail(self) -> dict:
        """Gets a dictionary of monitored and cached class/subclass pairs.
        
        Returns:
            `{ 'monitored': [(<class,subclass>)], 'cached': [<class,subclass)] }` 
        """
        _log.debug('Querying monitored/cached trace events')
        response = self.atcommand('AT%EVMON', filter=['%EVMON:'])
        if response[0] == 'ERROR':
            self._handle_at_error(response)
        response.remove('OK')
        detail = {
            'monitored': [],
            'cached': [],
        }
        if len(response) > 0:
            events: 'list[str]' = response[0].split(',')
            for event in events:
                trace_class = int(event.split('.')[0])
                trace_subclass = int(event.split('.')[1].replace('*', ''))
                detail['monitored'].append((trace_class, trace_subclass))
                if event.endswith('*'):
                    detail['cached'].append((trace_class, trace_subclass))
        return detail

    @property
    def trace_event_monitor(self) -> 'list[tuple[int, int]]':
        """The list of class/subclass pairs being monitored to cache."""
        return self._trace_detail()['monitored']
        
    @trace_event_monitor.setter
    def trace_event_monitor(self, events: 'list[tuple[int, int]]'):
        """Set a list of trace class/subclass pairs to monitor and cache."""
        command = 'AT%EVMON='
        for event in events:
            trace_class, trace_subclass = event
            if command != 'AT%EVMON=':
                command += ','
            command += f'{trace_class}.{trace_subclass}'
        _log.debug(f'Setting trace event monitoring for {events}')
        response = self.atcommand(command)
        if response[0] == 'ERROR':
            self._handle_at_error(response)

    @property
    def trace_events_cached(self) -> 'list[tuple[int, int]]':
        """The list of trace events cached for retrieval."""
        return self._trace_detail()['cached']

    def trace_event_get(self,
                        event: 'tuple[int, int]',
                        meta: bool = False,
                        ) -> 'str|dict':
        """Gets the cached event by class/subclass.

        Args:
            event: tuple of (class, subclass)
            meta: Returns the raw text string if False (default)
        
        Returns:
            String if meta is True, else metadata dictionary including:
            - `data_count` (int)
            - `signed_bitmask` (str)
            - `mobile_id` (str)
            - `timestamp` (str)
            - `class` (str)
            - `subclass` (str)
            - `priority` (str)
            - `data` (str)
        
        Raises:
            AtException

        """
        def signed32(n: int) -> int:
            """Converts an integer to signed 32-bit format."""
            n = n & 0xffffffff
            return (n ^ 0x80000000) - 0x80000000
        def event_timestamp(log_timestamp: int) -> int:
            offset = int(datetime(2001, 1, 1, tzinfo=timezone.utc).timestamp())
            return log_timestamp + offset
        if not (isinstance(event, tuple) and len(event) == 2):
            raise ValueError('event_get expects (class, subclass)')
        trace_class, trace_subclass = event
        _log.debug(f'Retrieving trace event class {trace_class}'
                   f' subclass {trace_subclass}')
        response = self.atcommand(f'AT%EVNT={trace_class},{trace_subclass}',
                                  filter=['%EVNT:'])
        #: res %EVNT: <dataCount>,<signedBitmask>,<MTID>,<timestamp>,
        # <class>,<subclass>,<priority>,<data0>,<data1>,..,<dataN>
        if response[0] == 'ERROR':
            self._handle_at_error(response)
        if not meta:
            return response[0]
        eventdata = response[0].split(',')
        event = {
            'data_count': int(eventdata[0]),
            'signed_bitmask': bin(int(eventdata[1])),
            'timestamp': event_timestamp(int(eventdata[3])),
            'class': int(eventdata[4]),
            'subclass': int(eventdata[5]),
            'priority': int(eventdata[6]),
            'raw_data': eventdata[7:],
            'data': {},
        }
        iso_time = datetime.utcfromtimestamp(event['timestamp']).isoformat()
        event['isotime'] = iso_time[:19] + 'Z'
        bitmask = event['signed_bitmask'][2:]
        while len(bitmask) < event['data_count']:
            bitmask = '0' + bitmask
        for i, bit in enumerate(reversed(bitmask)):
            if bit == '1':
                event['raw_data'][i] = signed32(int(event['raw_data'][i]))
            else:
                event['raw_data'][i] = int(event['raw_data'][i])
        # TODO lookup class/subclass definitions
        for trace_def in EVENT_TRACES:
            if trace_def.trace_class != trace_class:
                continue
            if trace_def.trace_subclass != trace_subclass:
                continue
            try:
                for i, value in enumerate(event['raw_data']):
                    tag, data_type = trace_def.data[i]
                    new_value = value
                    if 'flags' in tag and isinstance(data_type, dict):
                        new_value = []
                        for flag in data_type:
                            if flag & value:
                                new_value.append(data_type[flag])
                    elif str(tag).endswith('_state'):
                        if isinstance(data_type, dict):
                            new_value = data_type[value]
                        else:
                            try:   #: IntEnum
                                new_value = data_type(value)
                            except:
                                pass   # new_value stays as value
                    event['data'][tag] = new_value
            except Exception as err:
                _log.exception(err)
        return event

    @staticmethod
    def _list_events(bitmask: int) -> 'list[EventNotification]':
        events = []
        for notification in EventNotification:
            if bitmask & notification == notification:
                events.append(notification)
        return events

    @property
    def event_notification_monitor(self) -> 'list[EventNotification]':
        """The list of events monitored to assert the notification pin (`S88`)."""
        _log.debug('Querying event notification config')
        response = self.atcommand('ATS88?')
        if response[0] == 'ERROR':
            self._handle_at_error(response)
        return self._list_events(int(response[0]))
    
    @event_notification_monitor.setter
    def event_notification_monitor(self, event_list: 'list[EventNotification]'):
        bitmask = 0
        for event in event_list:
            bitmask = bitmask | event
        _log.debug(f'Setting event notifications: {event_list}')
        response = self.atcommand(f'ATS88={bitmask}')
        if response[0] == 'ERROR':
            self._handle_at_error(response)

    @property
    def event_notifications(self) -> 'list[EventNotification]':
        """The list of active events reported in `S89`."""
        _log.debug('Querying active event notifications')
        response = self.atcommand('ATS89?')
        if response[0] == 'ERROR':
            self._handle_at_error(response)
        return self._list_events(int(response[0]))
    
    @property
    def control_state(self) -> 'SatlliteControlState|None':
        """The control state enumerated value.
        
        Trace Class 3, Subclass 1, Data 22
        """
        self._satellite_status_get()
        if self._ctrl_state is not None:
            return SatlliteControlState(self._ctrl_state)
    
    @property
    def network_status(self) -> 'str|None':
        """The network status derived from control state."""
        if self._ctrl_state is None:
            self._satellite_status_get()
        if self._ctrl_state is not None:
            return SatlliteControlState(self._ctrl_state).name

    @property
    def registered(self) -> bool:
        """Indicates the modem is registered on the network."""
        return self.control_state == 10

    @property
    def beamsearch_state(self) -> 'BeamSearchState|None':
        """The beam search state (Trace Class 3, Subclass 1, Data 23)"""
        self._satellite_status_get()
        if self._beamsearch_state is not None:
            return BeamSearchState(self._beamsearch_state)
    
    @property
    def beamsearch(self) -> 'str|None':
        """The beam search state description."""
        if self._beamsearch_state is None:
            self._satellite_status_get()
        if self._beamsearch_state is None:
            return BeamSearchState(self._beamsearch_state).name

    @property
    def snr(self) -> 'float|None':
        """The average main beam Carrier-to-Noise (C/N0)."""
        self._satellite_status_get()
        return self._snr
            
    @property
    def signal_quality(self) -> SignalQuality:
        """Qualitative interpretation of the SNR."""
        signal_quality = SignalQuality.NONE
        snr = self.snr
        if snr is not None:
            if snr > SignalLevelRegional.INVALID.value:
                signal_quality = SignalQuality.WARNING
            elif snr > SignalLevelRegional.BARS_5.value:
                signal_quality = SignalQuality.STRONG
            elif snr > SignalLevelRegional.BARS_4.value:
                signal_quality = SignalQuality.GOOD
            elif snr > SignalLevelRegional.BARS_3.value:
                signal_quality = SignalQuality.MID
            elif snr > SignalLevelRegional.BARS_2.value:
                signal_quality = SignalQuality.LOW
            elif snr > SignalLevelRegional.BARS_1.value:
                signal_quality = SignalQuality.WEAK
        return signal_quality
    
    @property
    def satellite(self) -> 'str|None':
        """The current active satellite name."""
        if self._geo_beam_id is None:
            self._satellite_status_get()
        if self._geo_beam_id is not None:
            if GeoBeam.is_valid(self._geo_beam_id):
                return GeoBeam(self._geo_beam_id).satellite()
            return f'UNDEFINED {self._geo_beam_id}'
    
    @property
    def beam_id(self) -> 'str|None':
        """The current active regional beam ID of the active satellite."""
        if self._geo_beam_id is None:
            self._satellite_status_get()
        if self._geo_beam_id is not None:
            if GeoBeam.is_valid(self._geo_beam_id):
                return GeoBeam(self._geo_beam_id).id()
            return f'GEO{self._geo_beam_id}'
        
    def _satellite_status_get(self) -> None:
        if 'sat_status' in self._holdoffs:
            sat_status_age = int(time()) - self._holdoffs['sat_status']
            if sat_status_age < SAT_STATUS_HOLDOFF:
                _log.debug(f'Using cached status ({sat_status_age} s)')
                return
        else:
            sat_status_age = None
        _log.debug(f'Querying satellite status (cache {sat_status_age} s)')
        self._holdoffs['sat_status'] = int(time())
        # Trace events:
        #   Class 3 Subclass 1 C/N, Satellite Control State, Beam Search State
        #   Class 3 Subclass 5 Geo Beam ID
        command = ('ATS90=3 S91=1 S92=1 S116? S122? S123?'
                   ' S90=3 S91=5 S92=1 S102?')
        response = self.atcommand(command)
        if response[0] == 'ERROR':
            self._handle_at_error(response)
        self._snr = round(int(response[0]) / 100.0, 2)
        self._ctrl_state = int(response[1])
        self._beamsearch_state = int(response[2])
        try:   #: returns ERROR if no satellite has been acquired
            self._geo_beam_id = int(response[3])
        except:
            self._geo_beam_id = 0

    def satellite_status_get(self) -> dict:
        """Gets various satellite acquisition metrics.
        
        Returns:
            Dictionary including:
            - `satellite` (str)
            - `beam_id` (str)
            - `network_status` (str)
            - `control_state` (int)
            - `beamsearch` (str)
            - `beamsearch_state` (int)
            - `snr` (float)
        
        """
        return {
            'satellite': self.satellite,
            'beam_id': self.beam_id,
            'network_status': self.network_status,
            'control_state': self._ctrl_state,
            'beamsearch': self.beamsearch,
            'beamsearch_state': self._beamsearch_state,
            'snr': self._snr,
        }

    def shutdown(self) -> bool:
        """Tell the modem to prepare for power-down."""
        _log.warning('Attempting to shut down modem')
        response = self.atcommand('AT%OFF')
        if response[0] == 'ERROR':
            self._handle_at_error(response)
        return True

    @property
    def utc_time(self) -> str:
        """Gets current UTC time of the modem in ISO8601 format."""
        _log.debug('Querying system time')
        response = self.atcommand('AT%UTC', filter=['%UTC:'])
        if response[0] == 'ERROR':
            self._handle_at_error(response)
        return response[0].replace(' ', 'T') + 'Z'

    def s_register_get(self, register: 'str|int') -> int:
        """Gets the value of the S-register requested.

        Args:
            register: The register name/number (e.g. S80)

        Returns:
            integer value or None
        """
        if isinstance(register, str):
            try:
                register = int(register.replace('S', ''))
            except ValueError:
                raise ValueError(f'Invalid S-register {register}')
        _log.debug(f'Querying S-register {register}')
        response = self.atcommand(f'ATS{register}?')
        if response[0] == 'ERROR':
            self._handle_at_error(response)
        return int(response[0])

    def _s_registers_read(self) -> None:
        """Reads all defined S-registers."""
        command = 'AT'
        for reg in self.s_registers:
            if command != 'AT':
                command += ' '
            command += f'{reg}?'
        _log.debug('Querying all S-register values')
        response = self.atcommand(command)
        if response[0] == 'ERROR':
            _log.error('Could not read S-registers')
            raise
        index = 0
        for name, register in self.s_registers.items():
            register.value = response[index]
            index += 1

    def s_register_get_definitions(self) -> list:
        """(Future) Gets a list of S-register definitions.
        
        R=read-only, S=signed, V=volatile
        
        Returns:
            tuple(register, RSV, current, default, minimum, maximum) or None
        """
        raise NotImplementedError
        #: AT%SREG
        #: Sreg, RSV, CurrentVal, DefaultVal, MinimumVal, MaximumVal
        response = self.atcommand('AT%SREG')
        if response[0] == 'ERROR':
            self._handle_at_error(response)
        response.remove('OK')
        # header_rows = response[0:1]
        # Sreg RSV CurrentVal NvmValue DefaultValue MinimumValue MaximumVal
        reg_defs = response[2:]
        registers = []
        for row in reg_defs:
            reg_def = row.split(' ')
            reg_def = tuple(filter(None, reg_def))
            registers.append(reg_def)
        return registers



"""IDP Modem abstraction for pyserial-asyncio.

**WARNING**: This module is untested as a work in progress refactor of the
legacy atcommand_async module.

"""

import asyncio
import logging
import queue
from base64 import b64decode, b64encode
from datetime import datetime, timezone
from time import time

import serial_asyncio
from idpmodem.aterror import AtException, AtGnssTimeout
from idpmodem.constants import (AT_ERROR_CODES, EVENT_TRACES, GEOBEAMS,
                                POWER_MODES, WAKEUP_PERIODS, BeamSearchState,
                                DataFormat, EventNotification, MessagePriority,
                                MessageState, SatlliteControlState, TransmitterStatus)
from idpmodem.location import Location, location_from_nmea
from idpmodem.s_registers import SRegisters
from idpmodem.asyncio.atcommand import AtProtocol

GNSS_STALE_SECS = 1
GNSS_WAIT_SECS = 35
SAT_STATUS_HOLDOFF = 5

_log = logging.getLogger(__name__)


class ModemBusy(Exception):
    """Indicates the modem is busy processing a prior command."""


class AtConfiguration:
    def __init__(self) -> None:
        self.crc: bool = False
        self.echo: bool = True
        self.quiet: bool = False
        self.verbose: bool = True


class IdpModem:
    """A protocol factory abstracting AT commands for an IDP modem."""
    
    SERIAL_KWARGS = ['baudrate', 'timeout', 'write_timeout']
    BAUD_RATES = [1200, 2400, 4800, 9600, 19200]
    PROTOCOL_KWARGS = ['event_callback', 'at_timeout']
    OTHER_KWARGS = ['error_detail', 'debug', 'stale_secs', 'wait_secs']
    
    def __init__(self, serial_port: str, **kwargs):
        self.serial_kwargs = {
            'port': serial_port,
            'baudrate': int(kwargs.pop('baudrate', 9600)),
        }
        self.protocol_kwargs = {}
        self.error_detail = bool(kwargs.pop('error_detail', True))
        self.debug = bool(kwargs.pop('debug', False))
        for kwarg in kwargs:
            if kwarg in self.SERIAL_KWARGS:
                self.serial_kwargs[kwarg] = kwargs[kwarg]
            elif kwarg in self.PROTOCOL_KWARGS:
                self.protocol_kwargs[kwarg] = kwargs[kwarg]
        self.loop = asyncio.get_event_loop()
        self.coro = serial_asyncio.create_serial_connection(
            self.loop,
            protocol_factory=AtProtocol,
            **self.serial_kwargs
        )
        self.serial_port = None
        self.transport = None
        self.protocol = None
        self.commands = queue.Queue(1)
        self._at_config = AtConfiguration()
        self._mobile_id: str = None
        self._versions: dict = None
        self._power_mode: int = None
        self._wakeup_period: int = None
        self._ctrl_state: int = None
        self._snr: float = None
        self._geo_beam_id: int = None
        self._beamsearch_state: int = None
        self._loc_query: dict = {
            'stale_secs': int(kwargs.pop('stale_secs', GNSS_STALE_SECS)),
            'wait_secs': int(kwargs.pop('wait_secs', GNSS_WAIT_SECS)),
        }
        self._holdoffs: dict = {}   # used to ignore frequent repeat commands
        self._statistics: dict = {}
        self.s_registers = SRegisters()
        # self.tx_queue = queue.Queue()
        # self.tx_complete_callback: callable = None
        # self.rx_queue = queue.Queue()
        # self.rx_received_callback: callable = None
    
    def connect(self):
        self.transport, self.protocol = self.loop.run_until_complete(self.coro)
        self.loop.run_forever()
        _log.debug(f'Transport: {self.transport}')
        # self.serial_port = Serial(**self.serial_kwargs)

    def disconnect(self):
        self.loop.close()
        self.transport = None
        self.protocol = None
    
    @property
    def connected(self) -> bool:
        return self.transport is not None and self.protocol is not None

    @property
    def baudrate(self) -> 'int|None':
        return self.transport.baudrate if self.transport else None
    
    @baudrate.setter
    def baudrate(self, value: int):
        if not self.connected:
            raise ConnectionError('Modem is not connected')
        if value not in self.BAUD_RATES:
            raise ValueError(f'Baud rate must be one of {self.BAUD_RATES}')
        response = self.atcommand(f'AT+IPR={value}')
        if response and response[0] != 'ERROR':
            self.serial_port.baudrate = value

    @property
    def crc(self) -> 'bool|None':
        return self.protocol.crc if self.protocol is not None else None

    def atcommand(self,
                  command: str,
                  filter: 'list[str]' = [],
                  timeout: int = 5,
                  await_previous: bool = True,
                  ) -> 'list[str]':
        """Sends an AT command to the modem and returns the response.
        
        Args:
            command: The AT command
            filter: (optional) list of sub/strings to remove from response.
            timeout: Number of seconds to wait for a reply
                (not including messages queued by other threads)
            await_previous: If True, this will block if a prior command was
                submitted by another thread
        
        Returns:
            list of filtered and stripped response(s) to the command(s)
        
        Raises:
            ModemBusy if await_previous is False and a prior command is queued.
            AtException if an error occurred that is unrecognized.

        """
        if not self.connected:
            raise ConnectionError('No connection to IDP modem')
        while self.commands.full():
            if not await_previous:
                raise ModemBusy
            pass
        self.commands.put(command)
        # TODO: allow for async(?)
        res: list = self.protocol.command(command, filter, timeout, self.debug)
        if self.error_detail and res and res[0] == 'ERROR':
            _log.error(f'Error received for command {command}')
            err_res = self.protocol.command('ATS80?')
            if not err_res or err_res[0] == 'ERROR':
                raise AtException('Unhandled error getting last error code')
            last_err_code = err_res[0]
            detail = 'UNDEFINED'
            if int(last_err_code) in AT_ERROR_CODES:
                detail = AT_ERROR_CODES[int(last_err_code)]
            res.append(f'{detail} ({last_err_code})')
        self.commands.get()
        self.commands.task_done()
        return res
    
    def _handle_at_exception(self, response: 'list[str]') -> None:
        err = response[1] if self.error_detail else response[0]
        raise AtException(err)

    def config_init(self, crc: bool = False) -> bool:
        """Initializes modem communications with Echo, Verbose. CRC optional."""
        def attempt(command: str) -> bool:
            response = self.atcommand(command)
            return response[0] == 'OK'
        # try at most twice
        command = f'ATZ;E1;V1;Q0;%CRC={1 if crc else 0}'
        success = attempt(command)
        if not success:
            success = attempt(command)
        if success:
            self.protocol.crc = crc
            self._at_config.crc = crc
        return success

    def config_restore_nvm(self) -> bool:
        """Sends ATZ to restore config from non-volatile memory."""
        response = self.atcommand('ATZ')
        if response[0] == 'ERROR':
            return False
        return True

    def config_restore_factory(self) -> bool:
        """Sends AT&F to restore factory default and returns True on success."""
        response = self.atcommand('AT&F')
        if response[0] == 'ERROR':
            return False
        return True
    
    def config_report(self) -> 'tuple[dict, dict]':
        """Sends the AT&V command to retrieve S-register settings.
        
        Returns:
            A tuple with two dictionaries (empty if failed) with:
            at_config with booleans crc, echo, quiet and verbose
            reg_config with S-register tags and integer values
        
        Raises:
            AtException if an error was returned.

        """
        response = self.atcommand('AT&V')
        if response[0] == 'ERROR':
            self._handle_at_exception(response)
        at_config = response[1]
        s_regs = response[2]
        echo, quiet, verbose, crc = at_config.split(' ')
        self._at_config.crc = bool(int(crc[4]))
        self._at_config.echo = bool(int(echo[1]))
        self._at_config.quiet = bool(int(quiet[1]))
        self._at_config.verbose = bool(int(verbose[1]))
        reg_config = {}
        for reg in s_regs.split(' '):
            name, value = reg.split(':')
            reg_config[name] = int(value)
        return (at_config, reg_config)

    def config_volatile_report(self) -> 'dict|None':
        """Returns key S-register settings.
        
        GNSS Mode (S39), GNSS fix timeout (S41), GNSS Continuous (S55),
        GNSS Jamming Status (S56), GNSS Jamming Indicator (S57), 
        Low power Wakeup Period (S51)

        Returns:
            Dictionary of S-register values, or None if failed
            
        """
        register_list = [
            'S39',   #: GNSS Mode
            'S41',   #: GNSS Fix Timeout
            'S51',   #: Wakeup Interval
            'S55',   #: GNSS Continuous
            'S56',   #: GNSS Jamming Status
            'S57',   #: GNSS Jamming Indicator
        ]
        command = 'AT'
        for reg in register_list:
            command += f'{reg if command == "AT" else " " + reg}?'
        response = self.atcommand(command)
        if response[0] == 'ERROR':
            return None
        #: else
        response.remove('OK')
        volatile_regs = {}
        for r in range(len(response)):
            volatile_regs[register_list[r]] = int(response[r])
        return volatile_regs

    def config_nvm_save(self) -> bool:
        """Sends the AT&W command and returns True if successful."""
        response = self.atcommand('AT&W')
        return response[0] == 'OK'

    def crc_enable(self, enable: bool = True) -> bool:
        """Sends the AT%CRC command and returns success flag.
        
        Args:
            enable: turn on CRC if True else turn off

        Returns:
            True if the operation succeeded else False

        """
        command = f'AT%CRC={1 if enable else 0}'
        response = self.atcommand(command)
        if response[0] == 'ERROR':
            return False
        self.protocol.crc = enable
        self._at_config.crc = enable
        return True

    @property
    def mobile_id(self) -> 'str|None':
        """Returns the unique Mobile ID (Inmarsat serial number)."""
        if self._mobile_id is None:
            response = self.atcommand('AT+GSN', filter=['+GSN:'])
            if response[0] != 'ERROR':
                self._mobile_id = response[0]
        return self._mobile_id

    @property
    def versions(self) -> 'dict|None':
        """Returns the hardware, firmware and AT versions."""
        if not self._versions:
            response = self.atcommand('AT+GMR', filter=['+GMR:'])
            if response[0] != 'ERROR':
                fw_ver, hw_ver, at_ver = response[0].split(',')
                self._versions = {
                    'hardware': hw_ver,
                    'firmware': fw_ver,
                    'at': at_ver,
                }
        return self._versions

    @property
    def power_mode(self) -> 'str|None':
        if self._power_mode is None:
            response = self.atcommand('ATS50?')
            if response[0] != 'ERROR':
                self._power_mode = int(response[0])
        if self._power_mode in POWER_MODES:
            return POWER_MODES[self._power_mode]
    
    @power_mode.setter
    def power_mode(self, value: 'str|int'):
        if isinstance(value, str):
            if value not in POWER_MODES.values():
                raise ValueError(f'Invalid power mode {value}')
            for k, v in POWER_MODES.items():
                if v == value:
                    value = k
                    break
        if value not in POWER_MODES:
            raise ValueError(f'Invalid power mode {value}')
        response = self.atcommand(f'ATS50={value}')
        if response[0] == 'OK':
            self._power_mode = value
    
    @property
    def wakeup_period(self) -> 'str|None':
        if self._wakeup_period is None:
            response = self.atcommand('ATS51?')
            if response[0] != 'ERROR':
                self._wakeup_period = int(response[0])
        if self._wakeup_period in WAKEUP_PERIODS:
            return WAKEUP_PERIODS[self._wakeup_period]

    @wakeup_period.setter
    def wakeup_period(self, value: 'str|int'):
        if isinstance(value, str):
            if value not in WAKEUP_PERIODS.values():
                raise ValueError(f'Invalid wakeup period {value}')
            for k, v in WAKEUP_PERIODS.items():
                if v == value:
                    value = k
                    break
        if value not in WAKEUP_PERIODS:
            raise ValueError(f'Invalid wakeup period {value}')
        response = self.atcommand(f'ATS51={value}')
        if response[0] == 'OK':
            self._wakeup_period = value
    
    @property
    def gnss_refresh_interval(self) -> int:
        response = self.atcommand(f'ATS55?')
        if response[0] != 'ERROR':
            return int(response[0])

    @gnss_refresh_interval.setter
    def gnss_refresh_interval(self, value: int):
        self.gnss_continuous_set(value)

    def gnss_continuous_set(self,
                            interval: int = 0,
                            doppler: bool = True,
                            ) -> bool:
        """Sets the GNSS continous mode (0 = on-demand).
        
        Args:
            interval: Seconds between GNSS refresh.
            doppler: Often required for moving assets.
        
        Returns:
            True if successful setting.
        """
        if interval < 0 or interval > 30:
            raise ValueError('GNSS continuous interval must be in range 0..30')
        response = self.atcommand(f'AT%TRK={interval}{",1" if doppler else ""}')
        if response[0] == 'ERROR':
            return False
        return True

    def gnss_nmea_get(self,
                      stale_secs: int = GNSS_STALE_SECS,
                      wait_secs: int = GNSS_WAIT_SECS,
                      nmea: 'list[str]' = ['RMC', 'GSA', 'GGA', 'GSV'],
                      ) -> list:
        """Returns a list of NMEA-formatted sentences from GNSS.

        Args:
            stale_secs: Maximum age of fix in seconds (1..600)
            wait_secs: Maximum time to wait for fix (1..600)

        Returns:
            List of NMEA sentences

        Raises:
            ValueError if parameter out of range
            AtGnssTimeout if the fix timed out
            AtException if any other error code was returned

        """
        NMEA_SUPPORTED = ['RMC', 'GGA', 'GSA', 'GSV']
        BUFFER_SECONDS = 5
        if (stale_secs not in range(1, 600+1) or
            wait_secs not in range(1, 600+1)):
            raise ValueError('stale_secs and wait_secs must be 1..600')
        sentences = ''
        for sentence in nmea:
            sentence = sentence.upper()
            if sentence not in NMEA_SUPPORTED:
                raise ValueError(f'Unsupported NMEA sentence: {sentence}')
            if len(sentences) > 0:
                sentences += ','
            sentences += f'"{sentence}"'
        timeout = wait_secs + BUFFER_SECONDS
        request_time = time()
        response = self.atcommand(f'AT%GPS={stale_secs}'
                                        f',{wait_secs},{sentences}',
                                        timeout=timeout,
                                        filter=['%GPS:'])
        if response[0] == 'ERROR':
            if self.error_detail:
                if 'TIMEOUT' in response[1]:
                    raise AtGnssTimeout(response[1])
            self._handle_at_exception(response)
        response.remove('OK')
        time_to_fix = round(time() - request_time, 3)
        if 'gnss_ttf' not in self._statistics:
            self._statistics['gnss_ttf'] = time_to_fix
        else:
            old_ttf = self._statistics['gnss_ttf']
            avg_ttf = round((time_to_fix + old_ttf) / 2, 3)
            self._statistics['gnss_ttf'] = avg_ttf
        return response

    @property
    def location(self) -> 'Location|None':
        try:
            nmea_sentences = self.gnss_nmea_get(self._loc_query['stale_secs'],
                                                self._loc_query['wait_secs'])
            return location_from_nmea(nmea_sentences)
        except:
            return None

    def message_mo_send(self,
                        data: 'bytes|bytearray|str',
                        data_format: int = DataFormat.BASE64,
                        name: str = None,
                        priority: int = MessagePriority.LOW,
                        sin: int = None,
                        min: int = None,
                        ) -> str:
        """Submits a mobile-originated message to send.

        When submitting raw bytes, the first byte will be used as SIN. The
        first byte must not be in the reserved range (0..15).
        When submitting a string, the `sin` field is expected to be set and
        the data field will be appended to the `sin` byte and optionally the
        `min` byte if specified.
        
        Args:
            data: The data raw bytes or UTF-8 Text, Hexadecimal or Base64 string
            data_format: 1=text, 2=hexadecimal, 3=base64 (default)
            name: Optional unique name up to 8 characters long. If none is
                specified, use the 8 least-significant digits of unix timestamp.
            priority: 1=high, 4=low (default)
            sin: Optional first byte of payload used for codec, required if data
                is string type.
            min: Optional second byte of payload used for codec

        Returns:
            Name of the message if successful, or the error string.
        
        Raises:
            AtException if an error was returned by the modem.

        """
        name = str(int(time()))[-8:] if not name else name[0:8]
        if isinstance(data, bytes) or isinstance(data, bytearray):
            sin = data[0]
            data = b64encode(data[1:]).decode('utf-8')
            data_format = DataFormat.BASE64
        elif not isinstance(data, str):
            raise ValueError('Invalid data must be bytes, bytearray or string')
        elif not isinstance(sin, int) or sin not in range(16, 256):
            raise ValueError('Invalid SIN must be 16..255')
        if isinstance(min, int) and min not in range(0, 256):
            raise ValueError('Invalid MIN must be 0..255')
        min = f'.{min}' if min is not None else ''
        data = f'"{data}"' if data_format == DataFormat.TEXT else data
        command = f'AT%MGRT="{name}",{priority},{sin}{min},{data_format},{data}'
        response = self.atcommand(command)
        if response[0] == 'ERROR':
            self._handle_at_exception(response)
        return name

    def message_mo_state(self, name: str = None) -> 'list[dict]':
        """Returns the message state(s) requested.
        
        If no name filter is passed in, all available messages states
        are returned.  Returns False is the request failed.

        Args:
            name: The unique message name in the modem queue. If none is
                provided, all available message states in transmit queue will be
                returned.

        Returns:
            List of metadata for each message in transmit queue including:
            - `name` (str) The ID in the modem transmit queue
            - `state` (int) The state of the message
            - `state_name` (str) The `MessageState.name`
            - `size` (int) in bytes
            - `sent` (int) in bytes for large message progress

        """
        states = []
        name = f'="{name}"' if name is not None else ''
        response = self.atcommand(f'AT%MGRS{name}', filter=['%MGRS:'])
        # %MGRS: "<name>",<msg_no>,<priority>,<sin>,<state>,<size>,<sent_bytes>
        if response[0] != 'ERROR':
            response.remove('OK')
            for msg in response:
                if ',' in msg:
                    detail = msg.split(',')
                    states.append({
                        'name': detail[0].replace('"', ''),
                        'state': int(detail[4]),
                        'state_name': MessageState(int(detail[4])).name,
                        'size': int(detail[5]),
                        'sent': int(detail[6]),
                    })
        return states
    
    def message_mo_cancel(self, name: str) -> bool:
        """Cancels a mobile-originated message in the Tx ready state."""
        response = self.atcommand(f'AT%MGRC="{name}"')
        if response[0] == 'ERROR':
            return False
        return True

    def message_mo_clear(self) -> int:
        """Clears the modem transmit queue and returns the count cancelled.
        
        Returns:
            Count of messages deleted, or -1 in case of error

        """
        list_response = self.atcommand('AT%MGRL', filter=['%MGRL:'])
        if list_response[0] == 'ERROR':
            return -1
        message_count = len(list_response)
        for msg in list_response:
            del_response = self.atcommand(f'AT%MGRD={msg}C')
            if del_response[0] == 'ERROR':
                _log.error(f'Error clearing messages from transmit queue')
                return -1
        return message_count

    def message_mt_waiting(self) -> 'list[dict]':
        """Returns a list of received mobile-terminated message information.
        
        Returns:
            List of message metadata in the receive queue including:
            - `name` (str)
            - `sin` (int) first byte of payload
            - `priority` (int)
            - `state` (int) The state number
            - `state_name` (str) The `MessageState.name`
            - `size` (int) in bytes
            - `received` (int) in bytes for large message progress

        """
        waiting = []
        response = self.atcommand('AT%MGFN', filter=['%MGFN:'])
        #: %MGFN: "name",number,priority,sin,state,length,bytes_received
        if response[0] != 'ERROR':
            response.remove('OK')
            for msg in response:
                if (',' in msg):
                    detail = msg.split(',')
                    waiting.append({
                        'name': detail[0].replace('"', ''),
                        'sin': int(detail[3]),
                        'priority': int(detail[2]),
                        'state': int(detail[4]),
                        'state_name': MessageState(int(detail[4])).name,
                        'size': int(detail[5]),
                        'received': int(detail[6])
                        })
        return waiting

    def message_mt_get(self,
                       name: str,
                       data_format: int = DataFormat.BASE64,
                       meta: bool = False,
                       ) -> 'bytes|dict':
        """Returns the payload of a specified mobile-terminated message.
        
        Payload is presented as a string with encoding based on data_format. 

        Args:
            name: The unique name in the modem queue e.g. FM01.01
            data_format: text=1, hex=2, base64=3 (default)
            meta: If False returns raw bytes, else returns formatted data
                with metadata.

        Returns:
            The raw data bytes if meta is False, or a dictionary with:
            - `name` (str) The name assigned by the modem
            - `system_message_number` (int) System-assigned number
            - `system_message_sequence` (int) System-assigned number
            - `sin` (int) First byte of payload
            - `priority` (int)
            - `state` (int) The message state number
            - `state_name` (str) The `MessageState.name`
            - `size` (int) in bytes
            - `data_format` (int) 1=text, 2=hex, 3=base64
            - `data` (str) presented based on data_format

        """
        if not meta and data_format != DataFormat.BASE64:
            data_format = DataFormat.BASE64
        response = self.atcommand(f'AT%MGFG="{name}",{data_format}')
        if response[0] == 'ERROR':
            _log.error(f'Error retrieving message {name}')
            self._handle_at_exception(response)
        #: name, number, priority, sin, state, length, data_format, data
        try:
            detail = response[0].split(',')
            sys_msg_num, sys_msg_seq = detail[1].split('.')
            msg_sin = int(detail[3])
            data_str_no_sin = detail[7]
            if data_format == DataFormat.HEX:
                data = hex(msg_sin) + data_str_no_sin.lower()
            elif data_format == DataFormat.BASE64:
                # add SIN byte to base64 blob
                databytes = bytes([msg_sin]) + b64decode(data_str_no_sin)
                if not meta:
                    return databytes
                data = b64encode(databytes).decode('ascii')
            elif data_format == DataFormat.TEXT:
                data = f'\\{msg_sin:02x}' + data_str_no_sin
            return {
                'name': detail[0].replace('"', ''),
                'system_message_number': int(sys_msg_num),
                'system_message_sequence': int(sys_msg_seq),
                'priority': int(detail[2]),
                'sin': msg_sin,
                'state': int(detail[4]),
                'state_name': MessageState(int(detail[4])).name,
                'size': int(detail[5]),
                'data_format': data_format,
                'data': data
            }
        except Exception as err:
            _log.exception(err)

    def message_mt_delete(self, name: str) -> bool:
        """Marks a Return message for deletion by the modem.
        
        Args:
            name: The unique mobile-terminated name in the queue

        Returns:
            True if the operation succeeded

        """
        response = self.atcommand(f'AT%MGFM="{name}"')
        if response[0] == 'ERROR':
            err = f' ({response[1]})' if self.error_detail else ''
            _log.error(f'Error deleting message {name}{err}')
        return response[0] == 'OK'

    @property
    def transmitter_status(self):
        response = self.atcommand('ATS54?')
        if response[0] == 'ERROR':
            self._handle_at_exception(response)
        return TransmitterStatus(int(response[0]))

    def _trace_detail(self) -> dict:
        response = self.atcommand('AT%EVMON', filter=['%EVMON:'])
        if response[0] == 'ERROR':
            self._handle_at_exception(response)
        events = response[0].split(',')
        detail = {
            'monitored': [],
            'cached': [],
        }
        for event in events:
            trace_class = int(event.split('.')[0])
            trace_subclass = int(event.split('.')[1].replace('*', ''))
            detail['monitored'].append((trace_class, trace_subclass))
            if event.endswith('*'):
                detail['cached'].append((trace_class, trace_subclass))
        return detail

    @property
    def trace_event_monitor(self) -> 'list[(int, int)]':
        return self._trace_detail()['monitored']
        
    @trace_event_monitor.setter
    def trace_event_monitor(self, events: 'list[(int, int)]'):
        command = 'AT%EVMON='
        for event in events:
            trace_class, trace_subclass = event
            if command != 'AT%EVMON=':
                command += ','
            command += f'{trace_class}.{trace_subclass}'
        response = self.atcommand(command)
        if response[0] == 'ERROR':
            self._handle_at_exception(response)

    @property
    def trace_events_cached(self) -> list:
        return self._trace_detail()['cached']

    def trace_event_get(self,
                        event: 'tuple[int, int]',
                        meta: bool = False,
                        ) -> 'str|dict':
        """Gets the cached event by class/subclass.

        NOTE: Metadata feature is experimental.

        Args:
            event: tuple of (class, subclass)
            meta: Returns the raw text string if False (default)
        
        Returns:
            String if raw is True or metadata dictionary including:
            - `data_count` (int)
            - `signed_bitmask` (str)
            - `mobile_id` (str)
            - `timestamp` (str)
            - `class` (str)
            - `subclass` (str)
            - `priority` (str)
            - `data` (str)
        
        Raises:
            AtException

        """
        def signed32(n: int) -> int:
            """Converts an integer to signed 32-bit format."""
            n = n & 0xffffffff
            return (n ^ 0x80000000) - 0x80000000
        def event_timestamp(log_timestamp: int) -> int:
            offset = int(datetime(2001, 1, 1, tzinfo=timezone.utc).timestamp())
            return log_timestamp + offset
        if not (isinstance(event, tuple) and len(event) == 2):
            raise ValueError('event_get expects (class, subclass)')
        trace_class, trace_subclass = event
        response = self.atcommand(f'AT%EVNT={trace_class},{trace_subclass}',
                                  filter=['%EVNT:'])
        #: res %EVNT: <dataCount>,<signedBitmask>,<MTID>,<timestamp>,
        # <class>,<subclass>,<priority>,<data0>,<data1>,..,<dataN>
        if response[0] == 'ERROR':
            self._handle_at_exception(response)
        if not meta:
            return response[0]
        eventdata = response[0].split(',')
        event = {
            'data_count': int(eventdata[0]),
            'signed_bitmask': bin(int(eventdata[1])),
            'timestamp': event_timestamp(int(eventdata[3])),
            'class': int(eventdata[4]),
            'subclass': int(eventdata[5]),
            'priority': int(eventdata[6]),
            'data': eventdata[7:]
        }
        iso_time = datetime.utcfromtimestamp(event['timestamp']).isoformat()
        event['isotime'] = iso_time[:19] + 'Z'
        bitmask = event['signed_bitmask'][2:]
        while len(bitmask) < event['data_count']:
            bitmask = '0' + bitmask
        for i, bit in enumerate(reversed(bitmask)):
            if bit == '1':
                event['data'][i] = signed32(int(event['data'][i]))
            else:
                event['data'][i] = int(event['data'][i])
        # TODO lookup class/subclass definitions
        for trace_def in EVENT_TRACES:
            if trace_def.trace_class != trace_class:
                continue
            if trace_def.trace_subclass != trace_subclass:
                continue
            try:
                for i, value in enumerate(event['data']):
                    tag, data_type = trace_def.data[i]
                    new_value = value
                    if 'flags' in tag and isinstance(data_type, dict):
                        new_value = []
                        for flag in data_type:
                            if flag & value:
                                new_value.append(data_type[flag])
                    elif str(tag).endswith('_state'):
                        if isinstance(data_type, dict):
                            new_value = data_type[value]
                        else:
                            try:   #: IntEnum
                                new_value = data_type(value)
                            except:
                                pass   # new_value stays as value
                    event['data'][i] = { tag: new_value }
            except Exception as err:
                _log.exception(err)
        return event

    @staticmethod
    def _list_events(bitmask: int) -> 'list[EventNotification]':
        events = []
        for notification in EventNotification:
            if bitmask & notification == notification:
                events.append(notification)
        return events

    @property
    def event_notification_monitor(self) -> 'list[EventNotification]':
        response = self.atcommand('ATS88?')
        if response[0] == 'ERROR':
            self._handle_at_exception(response)
        return self._list_events(int(response[0]))
    
    @event_notification_monitor.setter
    def event_notification_monitor(self, event_list: 'list[EventNotification]'):
        bitmask = 0
        for event in event_list:
            bitmask = bitmask | event
        response = self.atcommand(f'ATS88={bitmask}')
        if response[0] == 'ERROR':
            self._handle_at_exception(response)

    @property
    def event_notifications(self) -> 'list[EventNotification]':
        response = self.atcommand('ATS89?')
        if response[0] == 'ERROR':
            self._handle_at_exception(response)
        return self._list_events(int(response[0]))
    
    @property
    def control_state(self) -> 'int|None':
        self.satellite_status_get()
        return SatlliteControlState(self._ctrl_state)
    
    @property
    def network_status(self) -> 'str|None':
        if self._ctrl_state is None:
            self.satellite_status_get()
        return SatlliteControlState(self._ctrl_state).name

    @property
    def beamsearch_state(self) -> 'int|None':
        self.satellite_status_get()
        return BeamSearchState(self._beamsearch_state)
    
    @property
    def beamsearch(self) -> 'str|None':
        if self._beamsearch_state is None:
            self.satellite_status_get()
        return BeamSearchState(self._beamsearch_state).name

    @property
    def snr(self) -> 'float|None':
        self.satellite_status_get()
        return self._snr
            
    @property
    def satellite(self) -> 'str|None':
        if self._geo_beam_id is None:
            self.satellite_status_get()
        if self._geo_beam_id is not None:
            if self._geo_beam_id in GEOBEAMS:
                return GEOBEAMS[self._geo_beam_id].split(' ')[0]
            return f'UNDEFINED {self._geo_beam_id}'
    
    @property
    def beam_id(self) -> 'str|None':
        if self._geo_beam_id is None:
            self.satellite_status_get()
        if self._geo_beam_id is not None:
            if self._geo_beam_id in GEOBEAMS:
                return GEOBEAMS[self._geo_beam_id].split(' ')[1]
            return f'GEO{self._geo_beam_id}'
        
    def satellite_status_get(self) -> dict:
        """Returns the control state and C/No.
        
        Returns:
            Dictionary including:
            - `satellite` (str)
            - `beam_id` (str)
            - `network_status` (str)
            - `control_state` (int)
            - `beamsearch` (str)
            - `beamsearch_state` (int)
            - `snr` (float)
        
        """
        if ('sat_status' in self._holdoffs and
            int(time()) - self._holdoffs['sat_status'] < SAT_STATUS_HOLDOFF):
            _log.debug('Ignoring repeat satellite status query')
            return
        _log.debug('Querying satellite status')
        self._holdoffs['sat_status'] = int(time())
        # Trace events:
        #   Class 3 Subclass 1 C/N, Satellite Control State, Beam Search State
        #   Class 3 Subclass 5 Geo Beam ID
        command = ('ATS90=3 S91=1 S92=1 S116? S122? S123?'
                   ' S90=3 S91=5 S92=1 S102?')
        response = self.atcommand(command)
        if response[0] == 'ERROR':
            self._handle_at_exception(response)
        self._snr = round(int(response[0]) / 100.0, 2)
        self._ctrl_state = int(response[1])
        self._beamsearch_state = int(response[2])
        self._geo_beam_id = int(response[3])
        return {
            'satellite': self.satellite,
            'beam_id': self.beam_id,
            'network_status': self.network_status,
            'control_state': self._ctrl_state,
            'beamsearch': self.beamsearch,
            'beamsearch_state': self._beamsearch_state,
            'snr': self._snr,
        }

    def shutdown(self) -> bool:
        """Tell the modem to prepare for power-down."""
        _log.warning('Attempting to shut down')
        response = self.atcommand('AT%OFF')
        if response[0] == 'ERROR':
            self._handle_at_exception(response)
        return True

    def utc_time(self) -> str:
        """Returns current UTC time of the modem in ISO8601 format."""
        _log.debug('Querying system time')
        response = self.atcommand('AT%UTC', filter=['%UTC:'])
        if response[0] == 'ERROR':
            self._handle_at_exception(response)
        return response[0].replace(' ', 'T') + 'Z'

    def s_register_get(self, register: 'str|int') -> int:
        """Returns the value of the S-register requested.

        Args:
            register: The register name/number (e.g. S80)

        Returns:
            integer value or None
        """
        if isinstance(register, str):
            try:
                register = int(register.replace('S', ''))
            except ValueError:
                raise ValueError(f'Invalid S-register {register}')
        _log.debug(f'Querying S-register {register}')
        response = self.atcommand(f'ATS{register}?')
        if response[0] == 'ERROR':
            self._handle_at_exception(response)
        return int(response[0])

    def _s_registers_read(self) -> None:
        command = 'AT'
        for reg in self.s_registers:
            if command != 'AT':
                command += ' '
            command += f'{reg}?'
        response = self.atcommand(command)
        if response[0] == 'ERROR':
            _log.error('Could not read S-registers')
            raise
        index = 0
        for name, register in self.s_registers.items():
            register.value = response[index]
            index += 1

    def s_register_get_definitions(self) -> list:
        """Returns a list of S-register definitions.
        R=read-only, S=signed, V=volatile
        
        Returns:
            tuple(register, RSV, current, default, minimum, maximum) or None
        """
        raise NotImplementedError
        #: AT%SREG
        #: Sreg, RSV, CurrentVal, DefaultVal, MinimumVal, MaximumVal
        response = self.atcommand('AT%SREG')
        if response[0] == 'ERROR':
            self._handle_at_exception(response)
        response.remove('OK')
        # header_rows = response[0:1]
        # Sreg RSV CurrentVal NvmValue DefaultValue MinimumValue MaximumVal
        reg_defs = response[2:]
        registers = []
        for row in reg_defs:
            reg_def = row.split(' ')
            reg_def = tuple(filter(None, reg_def))
            registers.append(reg_def)
        return registers
